"""Kernel Density Estimation Classifier."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging
import random
from enum import Enum

import torch
from torch import Tensor, nn

from anomalib.models.components import PCA, GaussianKDE

logger = logging.getLogger(__name__)


class FeatureScalingMethod(str, Enum):
    """Determines how the feature embeddings are scaled."""

    NORM = "norm"  # scale to unit vector length
    SCALE = "scale"  # scale to max length observed in training (preserve relative magnitude)


class KDEClassifier(nn.Module):
    """Classification module for KDE-based anomaly detection.

    Args:
        n_pca_components (int, optional): Number of PCA components. Defaults to 16.
        feature_scaling_method (FeatureScalingMethod, optional): Scaling method applied to features before passing to
            KDE. Options are `norm` (normalize to unit vector length) and `scale` (scale to max length observed in
            training).
        max_training_points (int, optional): Maximum number of training points to fit the KDE model. Defaults to 40000.
    """

    def __init__(
        self,
        n_pca_components: int = 16,
        feature_scaling_method: FeatureScalingMethod = FeatureScalingMethod.SCALE,
        max_training_points: int = 40000,
    ) -> None:
        super().__init__()

        self.n_pca_components = n_pca_components
        self.feature_scaling_method = feature_scaling_method
        self.max_training_points = max_training_points

        self.pca_model = PCA(n_components=self.n_pca_components)
        self.kde_model = GaussianKDE()

        self.register_buffer("max_length", torch.empty([]))
        self.max_length = torch.empty([])

    def pre_process(self, feature_stack: Tensor, max_length: Tensor | None = None) -> tuple[Tensor, Tensor]:
        """Pre-process the CNN features.

        Args:
          feature_stack (Tensor): Features extracted from CNN
          max_length (Tensor | None): Used to unit normalize the feature_stack vector. If ``max_len`` is not
            provided, the length is calculated from the ``feature_stack``. Defaults to None.

        Returns:
            (Tuple): Stacked features and length
        """

        if max_length is None:
            max_length = torch.max(torch.linalg.norm(feature_stack, ord=2, dim=1))

        if self.feature_scaling_method == FeatureScalingMethod.NORM:
            feature_stack /= torch.linalg.norm(feature_stack, ord=2, dim=1)[:, None]
        elif self.feature_scaling_method == FeatureScalingMethod.SCALE:
            feature_stack /= max_length
        else:
            raise RuntimeError("Unknown pre-processing mode. Available modes are: Normalized and Scale.")
        return feature_stack, max_length

    def fit(self, embeddings: Tensor) -> bool:
        """Fit a kde model to embeddings.

        Args:
            embeddings (Tensor): Input embeddings to fit the model.

        Returns:
            Boolean confirming whether the training is successful.
        """

        if embeddings.shape[0] < self.n_pca_components:
            logger.info("Not enough features to commit. Not making a model.")
            return False

        # if max training points is non-zero and smaller than number of staged features, select random subset
        if embeddings.shape[0] > self.max_training_points:
            selected_idx = torch.tensor(random.sample(range(embeddings.shape[0]), self.max_training_points))
            selected_features = embeddings[selected_idx]
        else:
            selected_features = embeddings

        feature_stack = self.pca_model.fit_transform(selected_features)
        feature_stack, max_length = self.pre_process(feature_stack)
        self.max_length = max_length
        self.kde_model.fit(feature_stack)

        return True

    def compute_kde_scores(self, features: Tensor, as_log_likelihood: bool | None = False) -> Tensor:
        """Compute the KDE scores.

        The scores calculated from the KDE model are converted to densities. If `as_log_likelihood` is set to true then
            the log of the scores are calculated.

        Args:
            features (Tensor): Features to which the PCA model is fit.
            as_log_likelihood (bool | None, optional): If true, gets log likelihood scores. Defaults to False.

        Returns:
            (Tensor): Score
        """

        features = self.pca_model.transform(features)
        features, _ = self.pre_process(features, self.max_length)
        # Scores are always assumed to be passed as a density
        kde_scores = self.kde_model(features)

        # add small constant to avoid zero division in log computation
        kde_scores += 1e-300

        if as_log_likelihood:
            kde_scores = torch.log(kde_scores)

        return kde_scores

    @staticmethod
    def compute_probabilities(scores: Tensor) -> Tensor:
        """Converts density scores to anomaly probabilities (see https://www.desmos.com/calculator/ifju7eesg7).

        Args:
          scores (Tensor): density of an image.

        Returns:
          probability that image with {density} is anomalous
        """
        return 1 / (1 + torch.exp(0.05 * (scores - 12)))

    def predict(self, features: Tensor) -> Tensor:
        """Predicts the probability that the features belong to the anomalous class.

        Args:
          features (Tensor): Feature from which the output probabilities are detected.

        Returns:
          Detection probabilities
        """

        scores = self.compute_kde_scores(features, as_log_likelihood=True)
        probabilities = self.compute_probabilities(scores)

        return probabilities

    def forward(self, features: Tensor) -> Tensor:
        """Make predictions on extracted features."""
        return self.predict(features)



"""Torch model defining the bottleneck layer."""

# Original Code
# Copyright (c) 2022 hq-deng
# https://github.com/hq-deng/RD4AD
# SPDX-License-Identifier: MIT
#
# Modified
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from __future__ import annotations

from typing import Callable

import torch
from torch import Tensor, nn
from torchvision.models.resnet import BasicBlock, Bottleneck


def conv3x3(in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1) -> nn.Conv2d:
    """3x3 convolution with padding."""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution."""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class OCBE(nn.Module):
    """One-Class Bottleneck Embedding module.

    Args:
        block (Bottleneck): Expansion value is extracted from this block.
        layers (int): Numbers of OCE layers to create after multiscale feature fusion.
        groups (int, optional): Number of blocked connections from input channels to output channels.
            Defaults to 1.
        width_per_group (int, optional): Number of layers in each intermediate convolution layer. Defaults to 64.
        norm_layer (Callable[..., nn.Module] | None, optional): Batch norm layer to use. Defaults to None.
    """

    def __init__(
        self,
        block: Bottleneck | BasicBlock,
        layers: int,
        groups: int = 1,
        width_per_group: int = 64,
        norm_layer: Callable[..., nn.Module] | None = None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer
        self.groups = groups
        self.base_width = width_per_group
        self.inplanes = 256 * block.expansion
        self.dilation = 1
        self.bn_layer = self._make_layer(block, 512, layers, stride=2)

        self.conv1 = conv3x3(64 * block.expansion, 128 * block.expansion, 2)
        self.bn1 = norm_layer(128 * block.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(128 * block.expansion, 256 * block.expansion, 2)
        self.bn2 = norm_layer(256 * block.expansion)
        self.conv3 = conv3x3(128 * block.expansion, 256 * block.expansion, 2)
        self.bn3 = norm_layer(256 * block.expansion)

        # self.conv4 and self.bn4 are from the original code:
        # https://github.com/hq-deng/RD4AD/blob/6554076872c65f8784f6ece8cfb39ce77e1aee12/resnet.py#L412
        self.conv4 = conv1x1(1024 * block.expansion, 512 * block.expansion, 1)
        self.bn4 = norm_layer(512 * block.expansion)

        for module in self.modules():
            if isinstance(module, nn.Conv2d):
                nn.init.kaiming_normal_(module.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(module, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(module.weight, 1)
                nn.init.constant_(module.bias, 0)

    def _make_layer(
        self,
        block: type[Bottleneck | BasicBlock],
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
    ) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes * 3, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes * 3,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def forward(self, features: list[Tensor]) -> Tensor:
        """Forward-pass of Bottleneck layer.

        Args:
            features (list[Tensor]): List of features extracted from the encoder.

        Returns:
            Tensor: Output of the bottleneck layer
        """
        # Always assumes that features has length of 3
        feature0 = self.relu(self.bn2(self.conv2(self.relu(self.bn1(self.conv1(features[0]))))))
        feature1 = self.relu(self.bn3(self.conv3(features[1])))
        feature_cat = torch.cat([feature0, feature1, features[2]], 1)
        output = self.bn_layer(feature_cat)
        # output = self.bn_layer(self.bn4(self.conv4(feature_cat)))

        return output.contiguous()


def get_bottleneck_layer(backbone: str, **kwargs) -> OCBE:
    """Get appropriate bottleneck layer based on the name of the backbone.

    Args:
        backbone (str): Name of the backbone.

    Returns:
        Bottleneck_layer: One-Class Bottleneck Embedding module.
    """
    if backbone in ("resnet18", "resnet34"):
        ocbe = OCBE(BasicBlock, 2, **kwargs)
    else:
        ocbe = OCBE(Bottleneck, 3, **kwargs)

    return ocbe



"""U-Flow: A U-shaped Normalizing Flow for Anomaly Detection with Unsupervised Threshold.

https://arxiv.org/pdf/2211.12353.pdf
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import torch
from omegaconf import DictConfig, ListConfig
from pytorch_lightning.callbacks import EarlyStopping
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch import Tensor

from anomalib.models.components import AnomalyModule
from anomalib.models.uflow.loss import UFlowLoss
from anomalib.models.uflow.torch_model import UflowModel

__all__ = ["Uflow", "UflowLightning"]


class Uflow(AnomalyModule):
    """PL Lightning Module for the UFLOW algorithm."""

    def __init__(
        self,
        input_size: tuple[int, int],
        backbone: str,
        flow_steps: int = 4,
        affine_clamp: float = 2.0,
        affine_subnet_channels_ratio: float = 1.0,
        permute_soft: bool = False,
    ) -> None:
        """
        Args:
            input_size (tuple[int, int]): Input image size.
            backbone (str): Backbone name.
            flow_steps (int): Number of flow steps.
            affine_clamp (float): Affine clamp.
            affine_subnet_channels_ratio (float): Affine subnet channels ratio.
            permute_soft (bool): Whether to use soft permutation.
        """
        super().__init__()
        self.model: UflowModel = UflowModel(
            input_size=input_size,
            backbone=backbone,
            flow_steps=flow_steps,
            affine_clamp=affine_clamp,
            affine_subnet_channels_ratio=affine_subnet_channels_ratio,
            permute_soft=permute_soft,
        )
        self.loss = UFlowLoss()

    def training_step(self, batch: dict[str, str | Tensor], *args, **kwargs) -> STEP_OUTPUT:
        z, ljd = self.model(batch["image"])
        loss = self.loss(z, ljd)
        self.log_dict({"loss": loss}, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        return {"loss": loss}

    def validation_step(self, batch: dict[str, str | Tensor], *args, **kwargs) -> STEP_OUTPUT:
        anomaly_maps = self.model(batch["image"])
        batch["anomaly_maps"] = anomaly_maps
        return batch


class UflowLightning(Uflow):
    """PL Lightning Module for the UFLOW algorithm.

    Args:
        hparams (DictConfig | ListConfig): Model params
    """

    def __init__(self, hparams: DictConfig | ListConfig) -> None:
        super().__init__(
            input_size=hparams.model.input_size,
            backbone=hparams.model.backbone,
            flow_steps=hparams.model.flow_steps,
            affine_clamp=hparams.model.affine_clamp,
            affine_subnet_channels_ratio=hparams.model.affine_subnet_channels_ratio,
            permute_soft=hparams.model.permute_soft,
        )
        self.lr = hparams.model.lr
        self.weight_decay = hparams.model.weight_decay
        self.hparams: DictConfig | ListConfig  # type: ignore
        self.save_hyperparameters(hparams)

    def configure_callbacks(self) -> list[EarlyStopping]:
        """Configure model-specific callbacks.

        Note:
            This method is used for the existing CLI.
            When PL CLI is introduced, configure callback method will be
                deprecated, and callbacks will be configured from either
                config.yaml file or from CLI.
        """
        early_stopping = EarlyStopping(
            monitor=self.hparams.model.early_stopping.metric,
            patience=self.hparams.model.early_stopping.patience,
            mode=self.hparams.model.early_stopping.mode,
        )
        return [early_stopping]

    def configure_optimizers(self):
        def get_total_number_of_iterations():
            return 25000

        # Optimizer
        optimizer = torch.optim.Adam(
            [{"params": self.parameters(), "initial_lr": self.lr}], lr=self.lr, weight_decay=self.weight_decay
        )

        # Scheduler for slowly reducing learning rate
        scheduler = torch.optim.lr_scheduler.LinearLR(
            optimizer, start_factor=1.0, end_factor=0.4, total_iters=get_total_number_of_iterations()
        )
        return [optimizer], [scheduler]



"""Test Config Getter."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import pytest

from anomalib.config import get_configurable_parameters


class TestConfig:
    """Test Config Getter."""

    def test_get_configurable_parameters_return_correct_model_name(self):
        """Configurable parameter should return the correct model name."""
        model_name = "stfpm"
        configurable_parameters = get_configurable_parameters(model_name)
        assert configurable_parameters.model.name == model_name

    def test_get_configurable_parameter_fails_with_none_arguments(self):
        """Configurable parameter should raise an error with none arguments."""
        with pytest.raises(ValueError):
            get_configurable_parameters()



"""Optical Flow extraction module for AI-VAD implementation."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import torch
import torchvision.transforms.functional as F
from torch import Tensor, nn
from torchvision.models.optical_flow import Raft_Large_Weights, raft_large


class FlowExtractor(nn.Module):
    """Optical Flow extractor.

    Computes the pixel displacement between 2 consecutive frames from a video clip.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        weights = Raft_Large_Weights.DEFAULT
        self.model = raft_large(weights=weights)
        self.transforms = weights.transforms()

    def pre_process(self, first_frame: Tensor, last_frame: Tensor) -> tuple[Tensor, Tensor]:
        """Resize inputs to dimensions required by backbone.

        Args:
            first_frame (Tensor): Starting frame of optical flow computation.
            last_frame (Tensor): Last frame of optical flow computation.
        Returns:
            tuple[Tensor, Tensor]: Preprocessed first and last frame.
        """
        first_frame = F.resize(first_frame, size=[520, 960], antialias=False)
        last_frame = F.resize(last_frame, size=[520, 960], antialias=False)
        return self.transforms(first_frame, last_frame)

    def forward(self, first_frame: Tensor, last_frame: Tensor) -> Tensor:
        """Forward pass through the flow extractor.

        Args:
            first_frame (Tensor): Batch of starting frames of shape (N, 3, H, W).
            last_frame (Tensor): Batch of last frames of shape (N, 3, H, W).
        Returns:
            Tensor: Estimated optical flow map of shape (N, 2, H, W).
        """
        height, width = first_frame.shape[-2:]

        # preprocess batch
        first_frame, last_frame = self.pre_process(first_frame, last_frame)

        # get flow maps
        with torch.no_grad():
            flows = self.model(first_frame, last_frame)[-1]

        # convert back to original size
        flows = F.resize(flows, [height, width], antialias=False)

        return flows



"""Helper functions for CFlow implementation."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging
import math

import numpy as np
import torch
from FrEIA.framework import SequenceINN
from torch import Tensor, nn

from anomalib.models.components.flow import AllInOneBlock

logger = logging.getLogger(__name__)


def get_logp(dim_feature_vector: int, p_u: Tensor, logdet_j: Tensor) -> Tensor:
    """Returns the log likelihood estimation.

    Args:
        dim_feature_vector (int): Dimensions of the condition vector
        p_u (Tensor): Random variable u
        logdet_j (Tensor): log of determinant of jacobian returned from the invertable decoder

    Returns:
        Tensor: Log probability
    """
    ln_sqrt_2pi = -np.log(np.sqrt(2 * np.pi))  # ln(sqrt(2*pi))
    logp = dim_feature_vector * ln_sqrt_2pi - 0.5 * torch.sum(p_u**2, 1) + logdet_j
    return logp


def positional_encoding_2d(condition_vector: int, height: int, width: int) -> Tensor:
    """Creates embedding to store relative position of the feature vector using sine and cosine functions.

    Args:
        condition_vector (int): Length of the condition vector
        height (int): H of the positions
        width (int): W of the positions

    Raises:
        ValueError: Cannot generate encoding with conditional vector length not as multiple of 4

    Returns:
        Tensor: condition_vector x HEIGHT x WIDTH position matrix
    """
    if condition_vector % 4 != 0:
        raise ValueError(f"Cannot use sin/cos positional encoding with odd dimension (got dim={condition_vector})")
    pos_encoding = torch.zeros(condition_vector, height, width)
    # Each dimension use half of condition_vector
    condition_vector = condition_vector // 2
    div_term = torch.exp(torch.arange(0.0, condition_vector, 2) * -(math.log(1e4) / condition_vector))
    pos_w = torch.arange(0.0, width).unsqueeze(1)
    pos_h = torch.arange(0.0, height).unsqueeze(1)
    pos_encoding[0:condition_vector:2, :, :] = (
        torch.sin(pos_w * div_term).transpose(0, 1).unsqueeze(1).repeat(1, height, 1)
    )
    pos_encoding[1:condition_vector:2, :, :] = (
        torch.cos(pos_w * div_term).transpose(0, 1).unsqueeze(1).repeat(1, height, 1)
    )
    pos_encoding[condition_vector::2, :, :] = (
        torch.sin(pos_h * div_term).transpose(0, 1).unsqueeze(2).repeat(1, 1, width)
    )
    pos_encoding[condition_vector + 1 :: 2, :, :] = (
        torch.cos(pos_h * div_term).transpose(0, 1).unsqueeze(2).repeat(1, 1, width)
    )
    return pos_encoding


def subnet_fc(dims_in: int, dims_out: int) -> nn.Sequential:
    """Subnetwork which predicts the affine coefficients.

    Args:
        dims_in (int): input dimensions
        dims_out (int): output dimensions

    Returns:
        nn.Sequential: Feed-forward subnetwork
    """
    return nn.Sequential(nn.Linear(dims_in, 2 * dims_in), nn.ReLU(), nn.Linear(2 * dims_in, dims_out))


def cflow_head(
    condition_vector: int, coupling_blocks: int, clamp_alpha: float, n_features: int, permute_soft: bool = False
) -> SequenceINN:
    """Create invertible decoder network.

    Args:
        condition_vector (int): length of the condition vector
        coupling_blocks (int): number of coupling blocks to build the decoder
        clamp_alpha (float): clamping value to avoid exploding values
        n_features (int): number of decoder features
        permute_soft (bool): Whether to sample the permutation matrix :math:`R` from :math:`SO(N)`,
            or to use hard permutations instead. Note, ``permute_soft=True`` is very slow
            when working with >512 dimensions.

    Returns:
        SequenceINN: decoder network block
    """
    coder = SequenceINN(n_features)
    logger.info("CNF coder: %d", n_features)
    for _ in range(coupling_blocks):
        coder.append(
            AllInOneBlock,
            cond=0,
            cond_shape=(condition_vector,),
            subnet_constructor=subnet_fc,
            affine_clamping=clamp_alpha,
            global_affine_type="SOFTPLUS",
            permute_soft=permute_soft,
        )
    return coder



"""Normality model of DFKDE."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from anomalib.models.components import FeatureExtractor
from anomalib.models.components.classification import (
    FeatureScalingMethod,
    KDEClassifier,
)

logger = logging.getLogger(__name__)


class DfkdeModel(nn.Module):
    """Normality Model for the DFKDE algorithm.

    Args:
        backbone (str): Pre-trained model backbone.
        pre_trained (bool, optional): Boolean to check whether to use a pre_trained backbone.
        n_comps (int, optional): Number of PCA components. Defaults to 16.
        pre_processing (str, optional): Preprocess features before passing to KDE.
            Options are between `norm` and `scale`. Defaults to "scale".
        filter_count (int, optional): Number of training points to fit the KDE model. Defaults to 40000.
        threshold_steepness (float, optional): Controls how quickly the value saturates around zero. Defaults to 0.05.
        threshold_offset (float, optional): Offset of the density function from 0. Defaults to 12.0.
    """

    def __init__(
        self,
        layers: list[str],
        backbone: str,
        pre_trained: bool = True,
        n_pca_components: int = 16,
        feature_scaling_method: FeatureScalingMethod = FeatureScalingMethod.SCALE,
        max_training_points: int = 40000,
    ) -> None:
        super().__init__()

        self.feature_extractor = FeatureExtractor(backbone=backbone, pre_trained=pre_trained, layers=layers).eval()

        self.classifier = KDEClassifier(
            n_pca_components=n_pca_components,
            feature_scaling_method=feature_scaling_method,
            max_training_points=max_training_points,
        )

    def get_features(self, batch: Tensor) -> Tensor:
        """Extract features from the pretrained network.

        Args:
            batch (Tensor): Image batch.

        Returns:
            Tensor: Tensor containing extracted features.
        """
        self.feature_extractor.eval()
        layer_outputs = self.feature_extractor(batch)
        for layer in layer_outputs:
            batch_size = len(layer_outputs[layer])
            layer_outputs[layer] = F.adaptive_avg_pool2d(input=layer_outputs[layer], output_size=(1, 1))
            layer_outputs[layer] = layer_outputs[layer].view(batch_size, -1)
        layer_outputs = torch.cat(list(layer_outputs.values())).detach()
        return layer_outputs

    def forward(self, batch: Tensor) -> Tensor:
        """Prediction by normality model.

        Args:
            batch (Tensor): Input images.

        Returns:
            Tensor: Predictions
        """

        # 1. apply feature extraction
        features = self.get_features(batch)
        if self.training:
            return features

        # 2. apply density estimation
        scores = self.classifier(features)
        return scores



__version__ = '1.0.1'
from .anomal_util.anomal_Tool import *
from .anomal_util.anomalib.config.config import *
from .anomal_util.anomalib.data.avenue import *
from .anomal_util.anomalib.data.btech import *
from .anomal_util.anomalib.data.folder import *
from .anomal_util.anomalib.data.folder_3d import *
from .anomal_util.anomalib.data.inference import *
from .anomal_util.anomalib.data.kolektor import *
from .anomal_util.anomalib.data.mvtec import *
from .anomal_util.anomalib.data.mvtec_3d import *
from .anomal_util.anomalib.data.shanghaitech import *
from .anomal_util.anomalib.data.synthetic import *
from .anomal_util.anomalib.data.task_type import *
from .anomal_util.anomalib.data.ucsd_ped import *
from .anomal_util.anomalib.data.visa import *
from .anomal_util.anomalib.data.base.datamodule import *
from .anomal_util.anomalib.data.base.dataset import *
from .anomal_util.anomalib.data.base.depth import *
from .anomal_util.anomalib.data.base.video import *
from .anomal_util.anomalib.data.utils.augmenter import *
from .anomal_util.anomalib.data.utils.boxes import *
from .anomal_util.anomalib.data.utils.download import *
from .anomal_util.anomalib.data.utils.image import *
from .anomal_util.anomalib.data.utils.label import *
from .anomal_util.anomalib.data.utils.path import *
from .anomal_util.anomalib.data.utils.split import *
from .anomal_util.anomalib.data.utils.transform import *
from .anomal_util.anomalib.data.utils.video import *
from .anomal_util.anomalib.data.utils.generators.perlin import *
from .anomal_util.anomalib.deploy.export import *
from .anomal_util.anomalib.deploy.inferencers.base_inferencer import *
from .anomal_util.anomalib.deploy.inferencers.openvino_inferencer import *
from .anomal_util.anomalib.deploy.inferencers.torch_inferencer import *
from .anomal_util.anomalib.models.ai_vad.density import *
from .anomal_util.anomalib.models.ai_vad.features import *
from .anomal_util.anomalib.models.ai_vad.flow import *
from .anomal_util.anomalib.models.ai_vad.lightning_model import *
from .anomal_util.anomalib.models.ai_vad.regions import *
from .anomal_util.anomalib.models.ai_vad.torch_model import *
from .anomal_util.anomalib.models.ai_vad.clip.clip import *
from .anomal_util.anomalib.models.ai_vad.clip.model import *
from .anomal_util.anomalib.models.cfa.anomaly_map import *
from .anomal_util.anomalib.models.cfa.lightning_model import *
from .anomal_util.anomalib.models.cfa.loss import *
from .anomal_util.anomalib.models.cfa.torch_model import *
from .anomal_util.anomalib.models.cflow.anomaly_map import *
from .anomal_util.anomalib.models.cflow.lightning_model import *
from .anomal_util.anomalib.models.cflow.torch_model import *
from .anomal_util.anomalib.models.cflow.utils import *
from .anomal_util.anomalib.models.components.base.anomaly_module import *
from .anomal_util.anomalib.models.components.base.dynamic_module import *
from .anomal_util.anomalib.models.components.classification.kde_classifier import *
from .anomal_util.anomalib.models.components.cluster.kmeans import *
from .anomal_util.anomalib.models.components.dimensionality_reduction.pca import *
from .anomal_util.anomalib.models.components.dimensionality_reduction.random_projection import *
from .anomal_util.anomalib.models.components.feature_extractors.timm import *
from .anomal_util.anomalib.models.components.feature_extractors.torchfx import *
from .anomal_util.anomalib.models.components.feature_extractors.utils import *
from .anomal_util.anomalib.models.components.filters.blur import *
from .anomal_util.anomalib.models.components.flow.all_in_one_block import *
from .anomal_util.anomalib.models.components.layers.sspcab import *
from .anomal_util.anomalib.models.components.sampling.k_center_greedy import *
from .anomal_util.anomalib.models.components.stats.kde import *
from .anomal_util.anomalib.models.components.stats.multi_variate_gaussian import *
from .anomal_util.anomalib.models.csflow.anomaly_map import *
from .anomal_util.anomalib.models.csflow.lightning_model import *
from .anomal_util.anomalib.models.csflow.loss import *
from .anomal_util.anomalib.models.csflow.torch_model import *
from .anomal_util.anomalib.models.dfkde.lightning_model import *
from .anomal_util.anomalib.models.dfkde.torch_model import *
from .anomal_util.anomalib.models.dfm.lightning_model import *
from .anomal_util.anomalib.models.dfm.torch_model import *
from .anomal_util.anomalib.models.draem.lightning_model import *
from .anomal_util.anomalib.models.draem.loss import *
from .anomal_util.anomalib.models.draem.torch_model import *
from .anomal_util.anomalib.models.dsr.anomaly_generator import *
from .anomal_util.anomalib.models.dsr.lightning_model import *
from .anomal_util.anomalib.models.dsr.loss import *
from .anomal_util.anomalib.models.dsr.torch_model import *
from .anomal_util.anomalib.models.efficient_ad.lightning_model import *
from .anomal_util.anomalib.models.efficient_ad.torch_model import *
from .anomal_util.anomalib.models.fastflow.anomaly_map import *
from .anomal_util.anomalib.models.fastflow.lightning_model import *
from .anomal_util.anomalib.models.fastflow.loss import *
from .anomal_util.anomalib.models.fastflow.torch_model import *
from .anomal_util.anomalib.models.ganomaly.lightning_model import *
from .anomal_util.anomalib.models.ganomaly.loss import *
from .anomal_util.anomalib.models.ganomaly.torch_model import *
from .anomal_util.anomalib.models.padim.anomaly_map import *
from .anomal_util.anomalib.models.padim.lightning_model import *
from .anomal_util.anomalib.models.padim.torch_model import *
from .anomal_util.anomalib.models.patchcore.anomaly_map import *
from .anomal_util.anomalib.models.patchcore.lightning_model import *
from .anomal_util.anomalib.models.patchcore.torch_model import *
from .anomal_util.anomalib.models.reverse_distillation.anomaly_map import *
from .anomal_util.anomalib.models.reverse_distillation.lightning_model import *
from .anomal_util.anomalib.models.reverse_distillation.loss import *
from .anomal_util.anomalib.models.reverse_distillation.torch_model import *
from .anomal_util.anomalib.models.reverse_distillation.components.bottleneck import *
from .anomal_util.anomalib.models.reverse_distillation.components.de_resnet import *
from .anomal_util.anomalib.models.rkde.feature_extractor import *
from .anomal_util.anomalib.models.rkde.lightning_model import *
from .anomal_util.anomalib.models.rkde.region_extractor import *
from .anomal_util.anomalib.models.rkde.torch_model import *
from .anomal_util.anomalib.models.stfpm.anomaly_map import *
from .anomal_util.anomalib.models.stfpm.lightning_model import *
from .anomal_util.anomalib.models.stfpm.loss import *
from .anomal_util.anomalib.models.stfpm.torch_model import *
from .anomal_util.anomalib.models.uflow.anomaly_map import *
from .anomal_util.anomalib.models.uflow.feature_extraction import *
from .anomal_util.anomalib.models.uflow.lightning_model import *
from .anomal_util.anomalib.models.uflow.loss import *
from .anomal_util.anomalib.models.uflow.torch_model import *
from .anomal_util.anomalib.post_processing.post_process import *
from .anomal_util.anomalib.post_processing.visualizer import *
from .anomal_util.anomalib.post_processing.normalization.cdf import *
from .anomal_util.anomalib.post_processing.normalization.min_max import *
from .anomal_util.anomalib.pre_processing.pre_process import *
from .anomal_util.anomalib.pre_processing.tiler import *
from .anomal_util.anomalib.pre_processing.transforms.custom import *
from .anomal_util.anomalib.utils.callbacks.cdf_normalization import *
from .anomal_util.anomalib.utils.callbacks.export import *
from .anomal_util.anomalib.utils.callbacks.graph import *
from .anomal_util.anomalib.utils.callbacks.metrics_configuration import *
from .anomal_util.anomalib.utils.callbacks.min_max_normalization import *
from .anomal_util.anomalib.utils.callbacks.model_loader import *
from .anomal_util.anomalib.utils.callbacks.post_processing_configuration import *
from .anomal_util.anomalib.utils.callbacks.tiler_configuration import *
from .anomal_util.anomalib.utils.callbacks.timer import *
from .anomal_util.anomalib.utils.callbacks.nncf.callback import *
from .anomal_util.anomalib.utils.callbacks.nncf.utils import *
from .anomal_util.anomalib.utils.callbacks.visualizer.visualizer_base import *
from .anomal_util.anomalib.utils.callbacks.visualizer.visualizer_image import *
from .anomal_util.anomalib.utils.callbacks.visualizer.visualizer_metric import *
from .anomal_util.anomalib.utils.cli.cli import *
from .anomal_util.anomalib.utils.cv.connected_components import *
from .anomal_util.anomalib.utils.hpo.config import *
from .anomal_util.anomalib.utils.hpo.runners import *
from .anomal_util.anomalib.utils.loggers.base import *
from .anomal_util.anomalib.utils.loggers.comet import *
from .anomal_util.anomalib.utils.loggers.tensorboard import *
from .anomal_util.anomalib.utils.loggers.wandb import *
from .anomal_util.anomalib.utils.metrics.anomaly_score_distribution import *
from .anomal_util.anomalib.utils.metrics.anomaly_score_threshold import *
from .anomal_util.anomalib.utils.metrics.aupr import *
from .anomal_util.anomalib.utils.metrics.aupro import *
from .anomal_util.anomalib.utils.metrics.auroc import *
from .anomal_util.anomalib.utils.metrics.binning import *
from .anomal_util.anomalib.utils.metrics.collection import *
from .anomal_util.anomalib.utils.metrics.min_max import *
from .anomal_util.anomalib.utils.metrics.optimal_f1 import *
from .anomal_util.anomalib.utils.metrics.plotting_utils import *
from .anomal_util.anomalib.utils.metrics.pro import *
from .anomal_util.anomalib.utils.sweep.config import *
from .anomal_util.anomalib.utils.sweep.helpers.callbacks import *
from .anomal_util.anomalib.utils.sweep.helpers.inference import *
from .anomal_util.tests.helpers.aupro_reference import *
from .anomal_util.tests.helpers.config import *
from .anomal_util.tests.helpers.dataset import *
from .anomal_util.tests.helpers.detection import *
from .anomal_util.tests.helpers.dummy import *
from .anomal_util.tests.helpers.inference import *
from .anomal_util.tests.helpers.metrics import *
from .anomal_util.tests.helpers.model import *
from .anomal_util.tests.helpers.shapes import *
from .anomal_util.tests.nightly.models.test_model_nightly import *
from .anomal_util.tests.nightly.tools.benchmarking.test_benchmarking import *
from .anomal_util.tests.pre_merge.config.test_config import *
from .anomal_util.tests.pre_merge.datasets.test_bounding_box_utils import *
from .anomal_util.tests.pre_merge.datasets.test_collate import *
from .anomal_util.tests.pre_merge.datasets.test_datamodule import *
from .anomal_util.tests.pre_merge.datasets.test_dataset import *
from .anomal_util.tests.pre_merge.datasets.test_synthetic_data import *
from .anomal_util.tests.pre_merge.datasets.test_tiler import *
from .anomal_util.tests.pre_merge.datasets.test_video import *
from .anomal_util.tests.pre_merge.deploy.test_inferencer import *
from .anomal_util.tests.pre_merge.models.test_feature_extractor import *
from .anomal_util.tests.pre_merge.models.test_model_premerge import *
from .anomal_util.tests.pre_merge.models.components.test_blur import *
from .anomal_util.tests.pre_merge.post_processing.test_visualizer import *
from .anomal_util.tests.pre_merge.pre_processing.test_tiler import *
from .anomal_util.tests.pre_merge.pre_processing.transforms.test_transforms import *
from .anomal_util.tests.pre_merge.tools.conftest import *
from .anomal_util.tests.pre_merge.tools.test_gradio_entrypoint import *
from .anomal_util.tests.pre_merge.tools.test_lightning_entrypoint import *
from .anomal_util.tests.pre_merge.tools.test_openvino_entrypoint import *
from .anomal_util.tests.pre_merge.tools.test_test_entrypoint import *
from .anomal_util.tests.pre_merge.tools.test_torch_entrypoint import *
from .anomal_util.tests.pre_merge.tools.test_train_entrypoint import *
from .anomal_util.tests.pre_merge.utils.test_config import *
from .anomal_util.tests.pre_merge.utils.callbacks.export_callback.dummy_lightning_model import *
from .anomal_util.tests.pre_merge.utils.callbacks.export_callback.test_export import *
from .anomal_util.tests.pre_merge.utils.callbacks.metrics_configuration_callback.test_metrics_configuration_callback import *
from .anomal_util.tests.pre_merge.utils.callbacks.normalization_callback.test_normalization_callback import *
from .anomal_util.tests.pre_merge.utils.callbacks.visualizer_callback.dummy_lightning_model import *
from .anomal_util.tests.pre_merge.utils.callbacks.visualizer_callback.test_visualizer import *
from .anomal_util.tests.pre_merge.utils.loggers.test_get_logger import *
from .anomal_util.tests.pre_merge.utils.metrics.test_adaptive_threshold import *
from .anomal_util.tests.pre_merge.utils.metrics.test_aupro import *
from .anomal_util.tests.pre_merge.utils.metrics.test_binning import *
from .anomal_util.tests.pre_merge.utils.metrics.test_pro import *
from .anomal_util.tests.pre_merge.utils.sweep.conftest import *
from .anomal_util.tests.pre_merge.utils.sweep.test_config import *
from .anomal_util.tests.pre_merge.utils.sweep.test_throughput import *
from .anomal_util.tools.test import *
from .anomal_util.tools.train import *
from .anomal_util.tools.benchmarking.benchmark import *
from .anomal_util.tools.benchmarking.utils.metrics import *
from .anomal_util.tools.hpo.sweep import *
from .anomal_util.tools.inference.gradio_inference import *
from .anomal_util.tools.inference.lightning_inference import *
from .anomal_util.tools.inference.openvino_inference import *
from .anomal_util.tools.inference.torch_inference import *
from .RT_DETR_util.RT_DETR_Tool import *
from .RT_DETR_util.src.core.config import *
from .RT_DETR_util.src.core.yaml_config import *
from .RT_DETR_util.src.core.yaml_utils import *
from .RT_DETR_util.src.data.dataloader import *
from .RT_DETR_util.src.data.functional import *
from .RT_DETR_util.src.data.transforms import *
from .RT_DETR_util.src.data.coco.coco_dataset import *
from .RT_DETR_util.src.data.coco.coco_eval import *
from .RT_DETR_util.src.data.coco.coco_utils import *
from .RT_DETR_util.src.misc.dist import *
from .RT_DETR_util.src.misc.logger import *
from .RT_DETR_util.src.misc.visualizer import *
from .RT_DETR_util.src.nn.arch.classification import *
from .RT_DETR_util.src.nn.backbone.common import *
from .RT_DETR_util.src.nn.backbone.presnet import *
from .RT_DETR_util.src.nn.backbone.test_resnet import *
from .RT_DETR_util.src.nn.backbone.utils import *
from .RT_DETR_util.src.nn.criterion.utils import *
from .RT_DETR_util.src.optim.amp import *
from .RT_DETR_util.src.optim.ema import *
from .RT_DETR_util.src.optim.optim import *
from .RT_DETR_util.src.solver.det_engine import *
from .RT_DETR_util.src.solver.det_solver import *
from .RT_DETR_util.src.solver.solver import *
from .RT_DETR_util.src.zoo.rtdetr.box_ops import *
from .RT_DETR_util.src.zoo.rtdetr.denoising import *
from .RT_DETR_util.src.zoo.rtdetr.hybrid_encoder import *
from .RT_DETR_util.src.zoo.rtdetr.matcher import *
from .RT_DETR_util.src.zoo.rtdetr.rtdetr import *
from .RT_DETR_util.src.zoo.rtdetr.rtdetr_criterion import *
from .RT_DETR_util.src.zoo.rtdetr.rtdetr_decoder import *
from .RT_DETR_util.src.zoo.rtdetr.rtdetr_postprocessor import *
from .RT_DETR_util.src.zoo.rtdetr.utils import *
from .RT_DETR_util.tools.export_onnx import *
from .RT_DETR_util.tools.train import *


"""PyTorch model for CFlow model implementation."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Any

import einops
import torch
from torch import nn

from anomalib.models.cflow.anomaly_map import AnomalyMapGenerator
from anomalib.models.cflow.utils import cflow_head, get_logp, positional_encoding_2d
from anomalib.models.components import FeatureExtractor


class CflowModel(nn.Module):
    """CFLOW: Conditional Normalizing Flows."""

    def __init__(
        self,
        input_size: tuple[int, int],
        backbone: str,
        layers: list[str],
        pre_trained: bool = True,
        fiber_batch_size: int = 64,
        decoder: str = "freia-cflow",
        condition_vector: int = 128,
        coupling_blocks: int = 8,
        clamp_alpha: float = 1.9,
        permute_soft: bool = False,
    ) -> None:
        super().__init__()

        self.backbone = backbone
        self.fiber_batch_size = fiber_batch_size
        self.condition_vector: int = condition_vector
        self.dec_arch = decoder
        self.pool_layers = layers

        self.encoder = FeatureExtractor(backbone=self.backbone, layers=self.pool_layers, pre_trained=pre_trained)
        self.pool_dims = self.encoder.out_dims
        self.decoders = nn.ModuleList(
            [
                cflow_head(
                    condition_vector=self.condition_vector,
                    coupling_blocks=coupling_blocks,
                    clamp_alpha=clamp_alpha,
                    n_features=pool_dim,
                    permute_soft=permute_soft,
                )
                for pool_dim in self.pool_dims
            ]
        )

        # encoder model is fixed
        for parameters in self.encoder.parameters():
            parameters.requires_grad = False

        self.anomaly_map_generator = AnomalyMapGenerator(image_size=input_size, pool_layers=self.pool_layers)

    def forward(self, images) -> Any:
        """Forward-pass images into the network to extract encoder features and compute probability.

        Args:
          images: Batch of images.

        Returns:
          Predicted anomaly maps.

        """

        self.encoder.eval()
        self.decoders.eval()
        with torch.no_grad():
            activation = self.encoder(images)

        distribution = [torch.Tensor(0).to(images.device) for _ in self.pool_layers]

        height: list[int] = []
        width: list[int] = []
        for layer_idx, layer in enumerate(self.pool_layers):
            encoder_activations = activation[layer]  # BxCxHxW

            batch_size, dim_feature_vector, im_height, im_width = encoder_activations.size()
            image_size = im_height * im_width
            embedding_length = batch_size * image_size  # number of rows in the conditional vector

            height.append(im_height)
            width.append(im_width)
            # repeats positional encoding for the entire batch 1 C H W to B C H W
            pos_encoding = einops.repeat(
                positional_encoding_2d(self.condition_vector, im_height, im_width).unsqueeze(0),
                "b c h w-> (tile b) c h w",
                tile=batch_size,
            ).to(images.device)
            c_r = einops.rearrange(pos_encoding, "b c h w -> (b h w) c")  # BHWxP
            e_r = einops.rearrange(encoder_activations, "b c h w -> (b h w) c")  # BHWxC
            decoder = self.decoders[layer_idx].to(images.device)

            # Sometimes during validation, the last batch E / N is not a whole number. Hence we need to add 1.
            # It is assumed that during training that E / N is a whole number as no errors were discovered during
            # testing. In case it is observed in the future, we can use only this line and ensure that FIB is at
            # least 1 or set `drop_last` in the dataloader to drop the last non-full batch.
            fiber_batches = embedding_length // self.fiber_batch_size + int(
                embedding_length % self.fiber_batch_size > 0
            )

            for batch_num in range(fiber_batches):  # per-fiber processing
                if batch_num < (fiber_batches - 1):
                    idx = torch.arange(batch_num * self.fiber_batch_size, (batch_num + 1) * self.fiber_batch_size)
                else:  # When non-full batch is encountered batch_num+1 * N will go out of bounds
                    idx = torch.arange(batch_num * self.fiber_batch_size, embedding_length)
                c_p = c_r[idx]  # NxP
                e_p = e_r[idx]  # NxC
                # decoder returns the transformed variable z and the log Jacobian determinant
                with torch.no_grad():
                    p_u, log_jac_det = decoder(e_p, [c_p])
                #
                decoder_log_prob = get_logp(dim_feature_vector, p_u, log_jac_det)
                log_prob = decoder_log_prob / dim_feature_vector  # likelihood per dim
                distribution[layer_idx] = torch.cat((distribution[layer_idx], log_prob))

        output = self.anomaly_map_generator(distribution=distribution, height=height, width=width)
        self.decoders.train()

        return output.to(images.device)



from setuptools import setup, find_packages

setup(
    name='fhailib',
    version='1.0.1',
    description='FH,FJ Vision AI SYSTEM LIB',
    author='LEE_DH',
    author_email='ldh9616@nate.com',
    url='https://github.com/LEE-pyt/AI_Module.git',
    install_requires=['tqdm', 'pandas', 'scikit-learn','torch==2.0.1','torchvision==0.15.2','onnx','onnxruntime','pycocotools','anomalib','GitPython','PyYAML','albumentations>=1.1.0','av>=10.0.0','comet-ml>=3.31.7','coverage[toml]','defusedxml==0.7.1','einops>=0.3.2','freia>=0.2','furo==2022.9.29','gitpython','gradio>=2.9.4','imgaug==0.4.0','ipykernel','ipywidgets','jsonargparse[signatures]>=4.3','kornia>=0.6.6,<0.6.10','matplotlib>=3.4.3','myst-parser','nbsphinx>=0.8.9','networkx~=2.5','nncf>=2.1.0','notebook','omegaconf>=2.1.1','onnx==1.14.0','onnxruntime==1.15.1','opencv-python>=4.5.3.56','openvino-dev>=2022.3.0','pandas>=1.1.0','pandoc','pre-commit','pycocotools','pytest','pytest-cov','pytest-order','pytest-sugar','pytest-xdist','pytorch-lightning>=1.7.0,<1.10.0','requests>=2.26.0','scipy','setuptools>=41.0.0','sphinx>=4.1.2','sphinx-autoapi','sphinxemoji==0.1.8','tensorboard','timm>=0.5.4,<=0.6.13','torch==2.0.1','torchmetrics==0.10.3','torchvision==0.15.2','tox','wandb==0.12.17'],
    packages=find_packages(exclude=[]),
    keywords=['FH_AI', 'FH_AI_TOOL', 'FH_AI_LIB', 'python tutorial', 'pypi'],
    python_requires='>=3.8',
    package_data={},
    zip_safe=False,
    py_modules=['AI_Module.anomal_util.anomal_Tool', 'AI_Module.anomal_util.anomalib.config.config', 'AI_Module.anomal_util.anomalib.data.avenue', 'AI_Module.anomal_util.anomalib.data.btech', 'AI_Module.anomal_util.anomalib.data.folder', 'AI_Module.anomal_util.anomalib.data.folder_3d', 'AI_Module.anomal_util.anomalib.data.inference', 'AI_Module.anomal_util.anomalib.data.kolektor', 'AI_Module.anomal_util.anomalib.data.mvtec', 'AI_Module.anomal_util.anomalib.data.mvtec_3d', 'AI_Module.anomal_util.anomalib.data.shanghaitech', 'AI_Module.anomal_util.anomalib.data.synthetic', 'AI_Module.anomal_util.anomalib.data.task_type', 'AI_Module.anomal_util.anomalib.data.ucsd_ped', 'AI_Module.anomal_util.anomalib.data.visa', 'AI_Module.anomal_util.anomalib.data.base.datamodule', 'AI_Module.anomal_util.anomalib.data.base.dataset', 'AI_Module.anomal_util.anomalib.data.base.depth', 'AI_Module.anomal_util.anomalib.data.base.video', 'AI_Module.anomal_util.anomalib.data.utils.augmenter', 'AI_Module.anomal_util.anomalib.data.utils.boxes', 'AI_Module.anomal_util.anomalib.data.utils.download', 'AI_Module.anomal_util.anomalib.data.utils.image', 'AI_Module.anomal_util.anomalib.data.utils.label', 'AI_Module.anomal_util.anomalib.data.utils.path', 'AI_Module.anomal_util.anomalib.data.utils.split', 'AI_Module.anomal_util.anomalib.data.utils.transform', 'AI_Module.anomal_util.anomalib.data.utils.video', 'AI_Module.anomal_util.anomalib.data.utils.generators.perlin', 'AI_Module.anomal_util.anomalib.deploy.export', 'AI_Module.anomal_util.anomalib.deploy.inferencers.base_inferencer', 'AI_Module.anomal_util.anomalib.deploy.inferencers.openvino_inferencer', 'AI_Module.anomal_util.anomalib.deploy.inferencers.torch_inferencer', 'AI_Module.anomal_util.anomalib.models.ai_vad.density', 'AI_Module.anomal_util.anomalib.models.ai_vad.features', 'AI_Module.anomal_util.anomalib.models.ai_vad.flow', 'AI_Module.anomal_util.anomalib.models.ai_vad.lightning_model', 'AI_Module.anomal_util.anomalib.models.ai_vad.regions', 'AI_Module.anomal_util.anomalib.models.ai_vad.torch_model', 'AI_Module.anomal_util.anomalib.models.ai_vad.clip.clip', 'AI_Module.anomal_util.anomalib.models.ai_vad.clip.model', 'AI_Module.anomal_util.anomalib.models.cfa.anomaly_map', 'AI_Module.anomal_util.anomalib.models.cfa.lightning_model', 'AI_Module.anomal_util.anomalib.models.cfa.loss', 'AI_Module.anomal_util.anomalib.models.cfa.torch_model', 'AI_Module.anomal_util.anomalib.models.cflow.anomaly_map', 'AI_Module.anomal_util.anomalib.models.cflow.lightning_model', 'AI_Module.anomal_util.anomalib.models.cflow.torch_model', 'AI_Module.anomal_util.anomalib.models.cflow.utils', 'AI_Module.anomal_util.anomalib.models.components.base.anomaly_module', 'AI_Module.anomal_util.anomalib.models.components.base.dynamic_module', 'AI_Module.anomal_util.anomalib.models.components.classification.kde_classifier', 'AI_Module.anomal_util.anomalib.models.components.cluster.kmeans', 'AI_Module.anomal_util.anomalib.models.components.dimensionality_reduction.pca', 'AI_Module.anomal_util.anomalib.models.components.dimensionality_reduction.random_projection', 'AI_Module.anomal_util.anomalib.models.components.feature_extractors.timm', 'AI_Module.anomal_util.anomalib.models.components.feature_extractors.torchfx', 'AI_Module.anomal_util.anomalib.models.components.feature_extractors.utils', 'AI_Module.anomal_util.anomalib.models.components.filters.blur', 'AI_Module.anomal_util.anomalib.models.components.flow.all_in_one_block', 'AI_Module.anomal_util.anomalib.models.components.layers.sspcab', 'AI_Module.anomal_util.anomalib.models.components.sampling.k_center_greedy', 'AI_Module.anomal_util.anomalib.models.components.stats.kde', 'AI_Module.anomal_util.anomalib.models.components.stats.multi_variate_gaussian', 'AI_Module.anomal_util.anomalib.models.csflow.anomaly_map','AI_Module.anomal_util.anomalib.models.csflow.lightning_model', 'AI_Module.anomal_util.anomalib.models.csflow.loss', 'AI_Module.anomal_util.anomalib.models.csflow.torch_model', 'AI_Module.anomal_util.anomalib.models.dfkde.lightning_model', 'AI_Module.anomal_util.anomalib.models.dfkde.torch_model', 'AI_Module.anomal_util.anomalib.models.dfm.lightning_model', 'AI_Module.anomal_util.anomalib.models.dfm.torch_model', 'AI_Module.anomal_util.anomalib.models.draem.lightning_model', 'AI_Module.anomal_util.anomalib.models.draem.loss', 'AI_Module.anomal_util.anomalib.models.draem.torch_model', 'AI_Module.anomal_util.anomalib.models.dsr.anomaly_generator', 'AI_Module.anomal_util.anomalib.models.dsr.lightning_model', 'AI_Module.anomal_util.anomalib.models.dsr.loss', 'AI_Module.anomal_util.anomalib.models.dsr.torch_model', 'AI_Module.anomal_util.anomalib.models.efficient_ad.lightning_model', 'AI_Module.anomal_util.anomalib.models.efficient_ad.torch_model', 'AI_Module.anomal_util.anomalib.models.fastflow.anomaly_map', 'AI_Module.anomal_util.anomalib.models.fastflow.lightning_model', 'AI_Module.anomal_util.anomalib.models.fastflow.loss', 'AI_Module.anomal_util.anomalib.models.fastflow.torch_model', 'AI_Module.anomal_util.anomalib.models.ganomaly.lightning_model', 'AI_Module.anomal_util.anomalib.models.ganomaly.loss', 'AI_Module.anomal_util.anomalib.models.ganomaly.torch_model', 'AI_Module.anomal_util.anomalib.models.padim.anomaly_map', 'AI_Module.anomal_util.anomalib.models.padim.lightning_model', 'AI_Module.anomal_util.anomalib.models.padim.torch_model', 'AI_Module.anomal_util.anomalib.models.patchcore.anomaly_map', 'AI_Module.anomal_util.anomalib.models.patchcore.lightning_model', 'AI_Module.anomal_util.anomalib.models.patchcore.torch_model', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.anomaly_map', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.lightning_model', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.loss', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.torch_model', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.components.bottleneck', 'AI_Module.anomal_util.anomalib.models.reverse_distillation.components.de_resnet', 'AI_Module.anomal_util.anomalib.models.rkde.feature_extractor', 'AI_Module.anomal_util.anomalib.models.rkde.lightning_model', 'AI_Module.anomal_util.anomalib.models.rkde.region_extractor', 'AI_Module.anomal_util.anomalib.models.rkde.torch_model', 'AI_Module.anomal_util.anomalib.models.stfpm.anomaly_map', 'AI_Module.anomal_util.anomalib.models.stfpm.lightning_model', 'AI_Module.anomal_util.anomalib.models.stfpm.loss', 'AI_Module.anomal_util.anomalib.models.stfpm.torch_model', 'AI_Module.anomal_util.anomalib.models.uflow.anomaly_map', 'AI_Module.anomal_util.anomalib.models.uflow.feature_extraction', 'AI_Module.anomal_util.anomalib.models.uflow.lightning_model', 'AI_Module.anomal_util.anomalib.models.uflow.loss', 'AI_Module.anomal_util.anomalib.models.uflow.torch_model', 'AI_Module.anomal_util.anomalib.post_processing.post_process', 'AI_Module.anomal_util.anomalib.post_processing.visualizer', 'AI_Module.anomal_util.anomalib.post_processing.normalization.cdf', 'AI_Module.anomal_util.anomalib.post_processing.normalization.min_max', 'AI_Module.anomal_util.anomalib.pre_processing.pre_process', 'AI_Module.anomal_util.anomalib.pre_processing.tiler', 'AI_Module.anomal_util.anomalib.pre_processing.transforms.custom', 'AI_Module.anomal_util.anomalib.utils.callbacks.cdf_normalization', 'AI_Module.anomal_util.anomalib.utils.callbacks.export', 'AI_Module.anomal_util.anomalib.utils.callbacks.graph', 'AI_Module.anomal_util.anomalib.utils.callbacks.metrics_configuration', 'AI_Module.anomal_util.anomalib.utils.callbacks.min_max_normalization', 'AI_Module.anomal_util.anomalib.utils.callbacks.model_loader', 'AI_Module.anomal_util.anomalib.utils.callbacks.post_processing_configuration', 'AI_Module.anomal_util.anomalib.utils.callbacks.tiler_configuration', 'AI_Module.anomal_util.anomalib.utils.callbacks.timer', 'AI_Module.anomal_util.anomalib.utils.callbacks.nncf.callback', 'AI_Module.anomal_util.anomalib.utils.callbacks.nncf.utils', 'AI_Module.anomal_util.anomalib.utils.callbacks.visualizer.visualizer_base', 'AI_Module.anomal_util.anomalib.utils.callbacks.visualizer.visualizer_image', 'AI_Module.anomal_util.anomalib.utils.callbacks.visualizer.visualizer_metric', 'AI_Module.anomal_util.anomalib.utils.cli.cli', 'AI_Module.anomal_util.anomalib.utils.cv.connected_components', 'AI_Module.anomal_util.anomalib.utils.hpo.config', 'AI_Module.anomal_util.anomalib.utils.hpo.runners', 'AI_Module.anomal_util.anomalib.utils.loggers.base', 'AI_Module.anomal_util.anomalib.utils.loggers.comet', 'AI_Module.anomal_util.anomalib.utils.loggers.tensorboard', 'AI_Module.anomal_util.anomalib.utils.loggers.wandb', 'AI_Module.anomal_util.anomalib.utils.metrics.anomaly_score_distribution', 'AI_Module.anomal_util.anomalib.utils.metrics.anomaly_score_threshold', 'AI_Module.anomal_util.anomalib.utils.metrics.aupr', 'AI_Module.anomal_util.anomalib.utils.metrics.aupro', 'AI_Module.anomal_util.anomalib.utils.metrics.auroc', 'AI_Module.anomal_util.anomalib.utils.metrics.binning', 'AI_Module.anomal_util.anomalib.utils.metrics.collection', 'AI_Module.anomal_util.anomalib.utils.metrics.min_max', 'AI_Module.anomal_util.anomalib.utils.metrics.optimal_f1', 'AI_Module.anomal_util.anomalib.utils.metrics.plotting_utils', 'AI_Module.anomal_util.anomalib.utils.metrics.pro', 'AI_Module.anomal_util.anomalib.utils.sweep.config', 'AI_Module.anomal_util.anomalib.utils.sweep.helpers.callbacks', 'AI_Module.anomal_util.anomalib.utils.sweep.helpers.inference', 'AI_Module.anomal_util.tests.helpers.aupro_reference', 'AI_Module.anomal_util.tests.helpers.config', 'AI_Module.anomal_util.tests.helpers.dataset', 'AI_Module.anomal_util.tests.helpers.detection', 'AI_Module.anomal_util.tests.helpers.dummy', 'AI_Module.anomal_util.tests.helpers.inference', 'AI_Module.anomal_util.tests.helpers.metrics', 'AI_Module.anomal_util.tests.helpers.model', 'AI_Module.anomal_util.tests.helpers.shapes', 'AI_Module.anomal_util.tests.nightly.models.test_model_nightly', 'AI_Module.anomal_util.tests.nightly.tools.benchmarking.test_benchmarking', 'AI_Module.anomal_util.tests.pre_merge.config.test_config', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_bounding_box_utils', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_collate', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_datamodule', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_dataset', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_synthetic_data', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_tiler', 'AI_Module.anomal_util.tests.pre_merge.datasets.test_video', 'AI_Module.anomal_util.tests.pre_merge.deploy.test_inferencer', 'AI_Module.anomal_util.tests.pre_merge.models.test_feature_extractor', 'AI_Module.anomal_util.tests.pre_merge.models.test_model_premerge', 'AI_Module.anomal_util.tests.pre_merge.models.components.test_blur', 'AI_Module.anomal_util.tests.pre_merge.post_processing.test_visualizer', 'AI_Module.anomal_util.tests.pre_merge.pre_processing.test_tiler', 'AI_Module.anomal_util.tests.pre_merge.pre_processing.transforms.test_transforms', 'AI_Module.anomal_util.tests.pre_merge.tools.conftest', 'AI_Module.anomal_util.tests.pre_merge.tools.test_gradio_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_lightning_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_openvino_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_test_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_torch_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.tools.test_train_entrypoint', 'AI_Module.anomal_util.tests.pre_merge.utils.test_config', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.export_callback.dummy_lightning_model', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.export_callback.test_export', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.metrics_configuration_callback.test_metrics_configuration_callback', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.normalization_callback.test_normalization_callback', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.visualizer_callback.dummy_lightning_model', 'AI_Module.anomal_util.tests.pre_merge.utils.callbacks.visualizer_callback.test_visualizer', 'AI_Module.anomal_util.tests.pre_merge.utils.loggers.test_get_logger', 'AI_Module.anomal_util.tests.pre_merge.utils.metrics.test_adaptive_threshold', 'AI_Module.anomal_util.tests.pre_merge.utils.metrics.test_aupro', 'AI_Module.anomal_util.tests.pre_merge.utils.metrics.test_binning', 'AI_Module.anomal_util.tests.pre_merge.utils.metrics.test_pro', 'AI_Module.anomal_util.tests.pre_merge.utils.sweep.conftest', 'AI_Module.anomal_util.tests.pre_merge.utils.sweep.test_config', 'AI_Module.anomal_util.tests.pre_merge.utils.sweep.test_throughput', 'AI_Module.anomal_util.tools.test', 'AI_Module.anomal_util.tools.train', 'AI_Module.anomal_util.tools.benchmarking.benchmark', 'AI_Module.anomal_util.tools.benchmarking.utils.metrics', 'AI_Module.anomal_util.tools.hpo.sweep', 'AI_Module.anomal_util.tools.inference.gradio_inference', 'AI_Module.anomal_util.tools.inference.lightning_inference', 'AI_Module.anomal_util.tools.inference.openvino_inference', 'AI_Module.anomal_util.tools.inference.torch_inference', 'AI_Module.RT_DETR_util.RT_DETR_Tool', 'AI_Module.RT_DETR_util.src.core.config', 'AI_Module.RT_DETR_util.src.core.yaml_config', 'AI_Module.RT_DETR_util.src.core.yaml_utils', 'AI_Module.RT_DETR_util.src.data.dataloader', 'AI_Module.RT_DETR_util.src.data.functional', 'AI_Module.RT_DETR_util.src.data.transforms', 'AI_Module.RT_DETR_util.src.data.coco.coco_dataset', 'AI_Module.RT_DETR_util.src.data.coco.coco_eval', 'AI_Module.RT_DETR_util.src.data.coco.coco_utils', 'AI_Module.RT_DETR_util.src.misc.dist', 'AI_Module.RT_DETR_util.src.misc.logger', 'AI_Module.RT_DETR_util.src.misc.visualizer', 'AI_Module.RT_DETR_util.src.nn.arch.classification', 'AI_Module.RT_DETR_util.src.nn.backbone.common', 'AI_Module.RT_DETR_util.src.nn.backbone.presnet', 'AI_Module.RT_DETR_util.src.nn.backbone.test_resnet', 'AI_Module.RT_DETR_util.src.nn.backbone.utils', 'AI_Module.RT_DETR_util.src.nn.criterion.utils', 'AI_Module.RT_DETR_util.src.optim.amp', 'AI_Module.RT_DETR_util.src.optim.ema', 'AI_Module.RT_DETR_util.src.optim.optim', 'AI_Module.RT_DETR_util.src.solver.det_engine', 'AI_Module.RT_DETR_util.src.solver.det_solver', 'AI_Module.RT_DETR_util.src.solver.solver', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.box_ops', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.denoising', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.hybrid_encoder', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.matcher', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.rtdetr', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.rtdetr_criterion', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.rtdetr_decoder', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.rtdetr_postprocessor', 'AI_Module.RT_DETR_util.src.zoo.rtdetr.utils', 'AI_Module.RT_DETR_util.tools.export_onnx', 'AI_Module.RT_DETR_util.tools.train'],

    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    package_dir={},
)



"""STFPM: Student-Teacher Feature Pyramid Matching for Unsupervised Anomaly Detection.

https://arxiv.org/abs/2103.04257
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import torch
from omegaconf import DictConfig, ListConfig
from pytorch_lightning.callbacks import EarlyStopping
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch import Tensor, optim

from anomalib.models.components import AnomalyModule
from anomalib.models.stfpm.loss import STFPMLoss
from anomalib.models.stfpm.torch_model import STFPMModel

__all__ = ["StfpmLightning"]


class Stfpm(AnomalyModule):
    """PL Lightning Module for the STFPM algorithm.

    Args:
        input_size (tuple[int, int]): Size of the model input.
        backbone (str): Backbone CNN network
        layers (list[str]): Layers to extract features from the backbone CNN
    """

    def __init__(
        self,
        input_size: tuple[int, int],
        backbone: str,
        layers: list[str],
    ) -> None:
        super().__init__()

        self.model = STFPMModel(
            input_size=input_size,
            backbone=backbone,
            layers=layers,
        )
        self.loss = STFPMLoss()

    def training_step(self, batch: dict[str, str | Tensor], *args, **kwargs) -> STEP_OUTPUT:
        """Training Step of STFPM.

        For each batch, teacher and student and teacher features are extracted from the CNN.

        Args:
          batch (dict[str, str | Tensor]): Input batch

        Returns:
          Loss value
        """
        del args, kwargs  # These variables are not used.

        self.model.teacher_model.eval()
        teacher_features, student_features = self.model.forward(batch["image"])
        loss = self.loss(teacher_features, student_features)
        self.log("train_loss", loss.item(), on_epoch=True, prog_bar=True, logger=True)
        return {"loss": loss}

    def validation_step(self, batch: dict[str, str | Tensor], *args, **kwargs) -> STEP_OUTPUT:
        """Validation Step of STFPM.

        Similar to the training step, student/teacher features are extracted from the CNN for each batch, and
        anomaly map is computed.

        Args:
          batch (dict[str, str | Tensor]): Input batch

        Returns:
          Dictionary containing images, anomaly maps, true labels and masks.
          These are required in `validation_epoch_end` for feature concatenation.
        """
        del args, kwargs  # These variables are not used.

        batch["anomaly_maps"] = self.model(batch["image"])
        return batch


class StfpmLightning(Stfpm):
    """PL Lightning Module for the STFPM algorithm.

    Args:
        hparams (DictConfig | ListConfig): Model params
    """

    def __init__(self, hparams: DictConfig | ListConfig) -> None:
        super().__init__(
            input_size=hparams.model.input_size,
            backbone=hparams.model.backbone,
            layers=hparams.model.layers,
        )
        self.hparams: DictConfig | ListConfig  # type: ignore
        self.save_hyperparameters(hparams)

    def configure_callbacks(self) -> list[EarlyStopping]:
        """Configure model-specific callbacks.

        Note:
            This method is used for the existing CLI.
            When PL CLI is introduced, configure callback method will be
                deprecated, and callbacks will be configured from either
                config.yaml file or from CLI.
        """
        early_stopping = EarlyStopping(
            monitor=self.hparams.model.early_stopping.metric,
            patience=self.hparams.model.early_stopping.patience,
            mode=self.hparams.model.early_stopping.mode,
        )
        return [early_stopping]

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """Configures optimizers.

        Note:
            This method is used for the existing CLI.
            When PL CLI is introduced, configure optimizers method will be
                deprecated, and optimizers will be configured from either
                config.yaml file or from CLI.

        Returns:
            Optimizer: SGD optimizer
        """
        return optim.SGD(
            params=self.model.student_model.parameters(),
            lr=self.hparams.model.lr,
            momentum=self.hparams.model.momentum,
            weight_decay=self.hparams.model.weight_decay,
        )



"""Loss function for the CS-Flow Model Implementation."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import torch
from torch import Tensor, nn


class CsFlowLoss(nn.Module):
    """Loss function for the CS-Flow Model Implementation."""

    def forward(self, z_dist: Tensor, jacobians: Tensor) -> Tensor:
        """Loss function of CS-Flow.

        Args:
            z_distribution (Tensor): Latent space image mappings from NF.
            jacobians (Tensor): Jacobians of the distribution

        Returns:
            Loss value
        """
        z_dist = torch.cat([z_dist[i].reshape(z_dist[i].shape[0], -1) for i in range(len(z_dist))], dim=1)
        return torch.mean(0.5 * torch.sum(z_dist**2, dim=(1,)) - jacobians) / z_dist.shape[1]



"""Fully Convolutional Cross-Scale-Flows for Image-based Defect Detection."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .lightning_model import Csflow, CsflowLightning

__all__ = ["Csflow", "CsflowLightning"]



"""Dataset Utils."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import warnings

import numpy as np
import torch
from torch import Tensor


class Denormalize:
    """Denormalize Torch Tensor into np image format."""

    def __init__(self, mean: list[float] | None = None, std: list[float] | None = None) -> None:
        """Denormalize Torch Tensor into np image format.

        Args:
            mean: Mean
            std: Standard deviation.
        """
        warnings.warn("Denormalize is no longer used and will be deprecated in v0.4.0")
        # If no mean and std provided, assign ImageNet values.
        if mean is None:
            mean = [0.485, 0.456, 0.406]

        if std is None:
            std = [0.229, 0.224, 0.225]

        self.mean = Tensor(mean)
        self.std = Tensor(std)

    def __call__(self, tensor: Tensor) -> np.ndarray:
        """Denormalize the input.

        Args:
            tensor (Tensor): Input tensor image (C, H, W)

        Returns:
            Denormalized numpy array (H, W, C).
        """
        if tensor.dim() == 4:
            if tensor.size(0):
                tensor = tensor.squeeze(0)
            else:
                raise ValueError(f"Tensor has batch size of {tensor.size(0)}. Only single batch is supported.")

        denormalized_per_channel = [(tnsr * std) + mean for tnsr, mean, std in zip(tensor, self.mean, self.std)]
        denormalized_tensor = torch.stack(denormalized_per_channel)

        denormalized_array = (denormalized_tensor * 255).permute(1, 2, 0).cpu().numpy().astype(np.uint8)
        return denormalized_array

    def __repr__(self) -> str:
        """Representational string."""
        return self.__class__.__name__ + "()"


class ToNumpy:
    """Convert Tensor into Numpy Array."""

    def __call__(self, tensor: Tensor, dims: tuple[int, ...] | None = None) -> np.ndarray:
        """Convert Tensor into Numpy Array.

        Args:
           tensor (Tensor): Tensor to convert. Input tensor in range 0-1.
           dims (tuple[int, ...] | None, optional): Convert dimensions from torch to numpy format.
                Tuple corresponding to axis permutation from torch tensor to numpy array. Defaults to None.

        Returns:
            Converted numpy ndarray.
        """
        # Default support is (C, H, W) or (N, C, H, W)
        if dims is None:
            dims = (0, 2, 3, 1) if len(tensor.shape) == 4 else (1, 2, 0)

        array = (tensor * 255).permute(dims).cpu().numpy().astype(np.uint8)

        if array.shape[0] == 1:
            array = array.squeeze(0)
        if array.shape[-1] == 1:
            array = array.squeeze(-1)

        return array

    def __repr__(self) -> str:
        """Representational string."""
        return self.__class__.__name__ + "()"



"""Dataset that generates synthetic anomalies.

This dataset can be used when there is a lack of real anomalous data.
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging
import math
import shutil
from copy import deepcopy
from pathlib import Path
from tempfile import mkdtemp

import albumentations as A
import cv2
import pandas as pd
from albumentations.pytorch import ToTensorV2
from pandas import DataFrame, Series

from anomalib.data.base.dataset import AnomalibDataset
from anomalib.data.task_type import TaskType
from anomalib.data.utils import Augmenter, Split, read_image

logger = logging.getLogger(__name__)


ROOT = "./.tmp/synthetic_anomaly"


def make_synthetic_dataset(
    source_samples: DataFrame, image_dir: Path, mask_dir: Path, anomalous_ratio: float = 0.5
) -> DataFrame:
    """Convert a set of normal samples into a mixed set of normal and synthetic anomalous samples.

    The synthetic images will be saved to the file system in the specified root directory under <root>/images.
    For the synthetic anomalous images, the masks will be saved under <root>/ground_truth.

    Args:
        source_samples (DataFrame): Normal images that will be used as source for the synthetic anomalous images.
        image_dir (Path): Directory to which the synthetic anomalous image files will be written.
        mask_dir (Path): Directory to which the ground truth anomaly masks will be written.
        anomalous_ratio (float): Fraction of source samples that will be converted into anomalous samples.
    """
    assert 1 not in source_samples.label_index.values, "All source images must be normal."
    assert image_dir.is_dir(), f"{image_dir} is not a folder."
    assert mask_dir.is_dir(), f"{mask_dir} is not a folder"

    # filter relevant columns
    source_samples = source_samples.filter(["image_path", "label", "label_index", "mask_path", "split"])
    # randomly select samples for augmentation
    n_anomalous = int(anomalous_ratio * len(source_samples))
    anomalous_samples = source_samples.sample(n_anomalous)
    normal_samples = source_samples.drop(anomalous_samples.index)
    anomalous_samples = anomalous_samples.reset_index(drop=True)

    # initialize augmenter
    augmenter = Augmenter("./datasets/dtd", p_anomalous=1.0, beta=(0.01, 0.2))

    # initialize transform for source images
    transform = A.Compose([A.ToFloat(), ToTensorV2()])

    def augment(sample: Series) -> Series:
        """Helper function to apply synthetic anomalous augmentation to a sample from a dataframe.

        Reads an image, applies the augmentations, writes the augmented image and corresponding mask to the file system,
        and returns a new Series object with the updates labels and file locations.

        Args:
            sample (Series): DataFrame row containing info about the image that will be augmented.

        Returns:
            Series: DataFrame row with updated information about the augmented image.
        """
        # read and transform image
        image = read_image(sample.image_path)
        image = transform(image=image)["image"].unsqueeze(0)
        # apply anomalous perturbation
        aug_im, mask = augmenter.augment_batch(image)
        # target file name with leading zeros
        file_name = f"{str(sample.name).zfill(int(math.log10(n_anomalous)) + 1)}.png"
        # write image
        aug_im = (aug_im.squeeze().permute((1, 2, 0)) * 255).numpy()
        aug_im = cv2.cvtColor(aug_im, cv2.COLOR_RGB2BGR)
        im_path = image_dir / file_name
        cv2.imwrite(str(im_path), aug_im)
        # write mask
        mask = (mask.squeeze() * 255).numpy()
        mask_path = mask_dir / file_name
        cv2.imwrite(str(mask_path), mask)
        out = dict(image_path=str(im_path), label="abnormal", label_index=1, mask_path=str(mask_path), split=Split.VAL)
        return Series(out)

    anomalous_samples = anomalous_samples.apply(augment, axis=1)

    samples = pd.concat([normal_samples, anomalous_samples], ignore_index=True)

    return samples


class SyntheticAnomalyDataset(AnomalibDataset):
    """Dataset which reads synthetically generated anomalous images from a temporary folder.

    Args:
        task (str): Task type, either "classification" or "segmentation".
        transform (A.Compose): Albumentations Compose object describing the transforms that are applied to the inputs.
        source_samples (DataFrame): Normal samples to which the anomalous augmentations will be applied.
    """

    def __init__(self, task: TaskType, transform: A.Compose, source_samples: DataFrame) -> None:
        super().__init__(task, transform)

        self.source_samples = source_samples

        # Files will be written to a temporary directory in the workdir, which is cleaned up after code execution
        root = Path(ROOT)
        root.mkdir(parents=True, exist_ok=True)

        self.root = Path(mkdtemp(dir=root))
        self.im_dir = self.root / "abnormal"
        self.mask_dir = self.root / "ground_truth"

        # create directories
        self.im_dir.mkdir()
        self.mask_dir.mkdir()

        self._cleanup = True  # flag that determines if temp dir is cleaned up when instance is deleted
        self.setup()

    @classmethod
    def from_dataset(cls, dataset: AnomalibDataset) -> SyntheticAnomalyDataset:
        """Create a synthetic anomaly dataset from an existing dataset of normal images.

        Args:
            dataset (AnomalibDataset): Dataset consisting of only normal images that will be converrted to a synthetic
                anomalous dataset with a 50/50 normal anomalous split.
        """
        return cls(task=dataset.task, transform=dataset.transform, source_samples=dataset.samples)

    def __copy__(self) -> SyntheticAnomalyDataset:
        """Returns a shallow copy of the dataset object and prevents cleanup when original object is deleted."""
        cls = self.__class__
        new = cls.__new__(cls)
        new.__dict__.update(self.__dict__)
        self._cleanup = False
        return new

    def __deepcopy__(self, _memo: dict) -> SyntheticAnomalyDataset:
        """Returns a deep copy of the dataset object and prevents cleanup when original object is deleted."""
        cls = self.__class__
        new = cls.__new__(cls)
        for key, value in self.__dict__.items():
            setattr(new, key, deepcopy(value))
        self._cleanup = False
        return new

    def _setup(self) -> None:
        """Create samples dataframe."""
        logger.info("Generating synthetic anomalous images for validation set")
        self.samples = make_synthetic_dataset(self.source_samples, self.im_dir, self.mask_dir, 0.5)

    def __del__(self) -> None:
        """Make sure the temporary directory is cleaned up when the dataset object is deleted."""
        if self._cleanup:
            shutil.rmtree(self.root)



"""Anomaly generator for the DSR model implementation."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from typing import List

import imgaug.augmenters as iaa
import numpy as np
import torch
from torch import Tensor, nn

from anomalib.data.utils.generators.perlin import _rand_perlin_2d_np


class DsrAnomalyGenerator(nn.Module):
    """Anomaly generator of the DSR model. The anomaly is generated using a Perlin
    noise generator on the two quantized representations of an image. This generator
    is only used during the second phase of training! The third phase requires generating
    smudges over the input images.

    Args:
        p_anomalous (float, optional): Probability to generate an anomalous image.
    """

    def __init__(
        self,
        p_anomalous: float = 0.5,
    ):
        super().__init__()

        self.p_anomalous = p_anomalous
        self.rot = iaa.Sequential([iaa.Affine(rotate=(-90, 90))])

    def generate_anomaly(self, height: int, width: int) -> Tensor:
        """Generate an anomalous mask.

        Args:
            height (int): Height of generated mask.
            width (int): Width of generated mask.

        Returns:
            Tensor: Generated mask.
        """
        min_perlin_scale = 0
        perlin_scale = 6
        perlin_scalex = 2 ** (torch.randint(min_perlin_scale, perlin_scale, (1,)).numpy()[0])
        perlin_scaley = 2 ** (torch.randint(min_perlin_scale, perlin_scale, (1,)).numpy()[0])
        threshold = 0.5
        perlin_noise_np = _rand_perlin_2d_np((height, width), (perlin_scalex, perlin_scaley))
        perlin_noise_np = self.rot(image=perlin_noise_np)
        mask = np.where(perlin_noise_np > threshold, np.ones_like(perlin_noise_np), np.zeros_like(perlin_noise_np))
        mask = np.expand_dims(mask, axis=2).astype(np.float32)
        mask = torch.from_numpy(mask)

        return mask

    def augment_batch(self, batch: Tensor) -> Tensor:
        """Generate anomalous augmentations for a batch of input images.

        Args:
            batch (Tensor): Batch of input images

        Returns:
            Tensor: Ground truth masks corresponding to the anomalous perturbations.
        """
        batch_size, _, height, width = batch.shape

        # Collect perturbations
        masks_list: List[Tensor] = []
        for _ in range(batch_size):
            if torch.rand(1) > self.p_anomalous:  # include normal samples
                masks_list.append(torch.zeros((1, height, width)))
            else:
                mask = self.generate_anomaly(height, width)
                masks_list.append(mask.permute((2, 0, 1)))

        masks = torch.stack(masks_list).to(batch.device)

        return masks



"""Feature Extractor.

This script extracts features from a CNN network
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging
import warnings

import timm
import torch
from torch import Tensor, nn

logger = logging.getLogger(__name__)


class TimmFeatureExtractor(nn.Module):
    """Extract features from a CNN.

    Args:
        backbone (nn.Module): The backbone to which the feature extraction hooks are attached.
        layers (Iterable[str]): List of layer names of the backbone to which the hooks are attached.
        pre_trained (bool): Whether to use a pre-trained backbone. Defaults to True.
        requires_grad (bool): Whether to require gradients for the backbone. Defaults to False.
            Models like ``stfpm`` use the feature extractor model as a trainable network. In such cases gradient
            computation is required.

    Example:
        >>> import torch
        >>> from anomalib.models.components.feature_extractors import TimmFeatureExtractor

        >>> model = TimmFeatureExtractor(model="resnet18", layers=['layer1', 'layer2', 'layer3'])
        >>> input = torch.rand((32, 3, 256, 256))
        >>> features = model(input)

        >>> [layer for layer in features.keys()]
            ['layer1', 'layer2', 'layer3']
        >>> [feature.shape for feature in features.values()]
            [torch.Size([32, 64, 64, 64]), torch.Size([32, 128, 32, 32]), torch.Size([32, 256, 16, 16])]
    """

    def __init__(self, backbone: str, layers: list[str], pre_trained: bool = True, requires_grad: bool = False):
        super().__init__()

        # Extract backbone-name and weight-URI from the backbone string.
        if "__AT__" in backbone:
            backbone, uri = backbone.split("__AT__")
            pretrained_cfg = timm.models.registry.get_pretrained_cfg(backbone)
            # Override pretrained_cfg["url"] to use different pretrained weights.
            pretrained_cfg["url"] = uri
        else:
            pretrained_cfg = None

        self.backbone = backbone
        self.layers = layers
        self.idx = self._map_layer_to_idx()
        self.requires_grad = requires_grad
        self.feature_extractor = timm.create_model(
            backbone,
            pretrained=pre_trained,
            pretrained_cfg=pretrained_cfg,
            features_only=True,
            exportable=True,
            out_indices=self.idx,
        )
        self.out_dims = self.feature_extractor.feature_info.channels()
        self._features = {layer: torch.empty(0) for layer in self.layers}

    def _map_layer_to_idx(self, offset: int = 3) -> list[int]:
        """Maps set of layer names to indices of model.

        Args:
            offset (int) `timm` ignores the first few layers when indexing please update offset based on need

        Returns:
            Feature map extracted from the CNN
        """
        idx = []
        features = timm.create_model(
            self.backbone,
            pretrained=False,
            features_only=False,
            exportable=True,
        )
        for i in self.layers:
            try:
                idx.append(list(dict(features.named_children()).keys()).index(i) - offset)
            except ValueError:
                warnings.warn(f"Layer {i} not found in model {self.backbone}")
                # Remove unfound key from layer dict
                self.layers.remove(i)

        return idx

    def forward(self, inputs: Tensor) -> dict[str, Tensor]:
        """Forward-pass input tensor into the CNN.

        Args:
            inputs (Tensor): Input tensor

        Returns:
            Feature map extracted from the CNN
        """
        if self.requires_grad:
            features = dict(zip(self.layers, self.feature_extractor(inputs)))
        else:
            self.feature_extractor.eval()
            with torch.no_grad():
                features = dict(zip(self.layers, self.feature_extractor(inputs)))
        return features


class FeatureExtractor(TimmFeatureExtractor):
    """Compatibility wrapper for the old FeatureExtractor class.

    See :class:`anomalib.models.components.feature_extractors.timm.TimmFeatureExtractor` for more details.
    """

    def __init__(self, *args, **kwargs):
        logger.warning(
            "FeatureExtractor is deprecated. Use TimmFeatureExtractor instead."
            " Both FeatureExtractor and TimmFeatureExtractor will be removed in a future release."
        )
        super().__init__(*args, **kwargs)



"""Tools for anomaly score normalization."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class NormalizationMethod(str, Enum):
    """Normalization method for normalization."""

    CDF = "cdf"
    MIN_MAX = "min_max"
    NONE = "none"



"""by lyuwenyu
"""

import math
import torch 
import torch.nn as nn
import torch.nn.functional as F 


def inverse_sigmoid(x: torch.Tensor, eps: float=1e-5) -> torch.Tensor:
    x = x.clip(min=0., max=1.)
    return torch.log(x.clip(min=eps) / (1 - x).clip(min=eps))


def deformable_attention_core_func(value, value_spatial_shapes, sampling_locations, attention_weights):
    """
    Args:
        value (Tensor): [bs, value_length, n_head, c]
        value_spatial_shapes (Tensor|List): [n_levels, 2]
        value_level_start_index (Tensor|List): [n_levels]
        sampling_locations (Tensor): [bs, query_length, n_head, n_levels, n_points, 2]
        attention_weights (Tensor): [bs, query_length, n_head, n_levels, n_points]

    Returns:
        output (Tensor): [bs, Length_{query}, C]
    """
    bs, _, n_head, c = value.shape
    _, Len_q, _, n_levels, n_points, _ = sampling_locations.shape

    split_shape = [h * w for h, w in value_spatial_shapes]
    value_list = value.split(split_shape, dim=1)
    sampling_grids = 2 * sampling_locations - 1
    sampling_value_list = []
    for level, (h, w) in enumerate(value_spatial_shapes):
        # N_, H_*W_, M_, D_ -> N_, H_*W_, M_*D_ -> N_, M_*D_, H_*W_ -> N_*M_, D_, H_, W_
        value_l_ = value_list[level].flatten(2).permute(
            0, 2, 1).reshape(bs * n_head, c, h, w)
        # N_, Lq_, M_, P_, 2 -> N_, M_, Lq_, P_, 2 -> N_*M_, Lq_, P_, 2
        sampling_grid_l_ = sampling_grids[:, :, :, level].permute(
            0, 2, 1, 3, 4).flatten(0, 1)
        # N_*M_, D_, Lq_, P_
        sampling_value_l_ = F.grid_sample(
            value_l_,
            sampling_grid_l_,
            mode='bilinear',
            padding_mode='zeros',
            align_corners=False)
        sampling_value_list.append(sampling_value_l_)
    # (N_, Lq_, M_, L_, P_) -> (N_, M_, Lq_, L_, P_) -> (N_*M_, 1, Lq_, L_*P_)
    attention_weights = attention_weights.permute(0, 2, 1, 3, 4).reshape(
        bs * n_head, 1, Len_q, n_levels * n_points)
    output = (torch.stack(
        sampling_value_list, dim=-2).flatten(-2) *
              attention_weights).sum(-1).reshape(bs, n_head * c, Len_q)

    return output.permute(0, 2, 1)


import math 
def bias_init_with_prob(prior_prob=0.01):
    """initialize conv/fc bias value according to a given probability value."""
    bias_init = float(-math.log((1 - prior_prob) / prior_prob))
    return bias_init



def get_activation(act: str, inpace: bool=True):
    '''get activation
    '''
    act = act.lower()
    
    if act == 'silu':
        m = nn.SiLU()

    elif act == 'relu':
        m = nn.ReLU()

    elif act == 'leaky_relu':
        m = nn.LeakyReLU()

    elif act == 'silu':
        m = nn.SiLU()
    
    elif act == 'gelu':
        m = nn.GELU()
        
    elif act is None:
        m = nn.Identity()
    
    elif isinstance(act, nn.Module):
        m = act

    else:
        raise RuntimeError('')  

    if hasattr(m, 'inplace'):
        m.inplace = inpace
    
    return m 





"""Anomaly Map Generator for CS-Flow model."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from enum import Enum

import torch
import torch.nn.functional as F
from torch import Tensor, nn


class AnomalyMapMode(str, Enum):
    """Generate anomaly map from all the scales or the max."""

    ALL = "all"
    MAX = "max"


class AnomalyMapGenerator(nn.Module):
    """Anomaly Map Generator for CS-Flow model.

    Args:
        input_dims (tuple[int, int, int]): Input dimensions.
        mode (AnomalyMapMode): Anomaly map mode. Defaults to AnomalyMapMode.ALL.
    """

    def __init__(self, input_dims: tuple[int, int, int], mode: AnomalyMapMode = AnomalyMapMode.ALL) -> None:
        super().__init__()
        self.mode = mode
        self.input_dims = input_dims

    def forward(self, inputs: Tensor) -> Tensor:
        """Get anomaly maps by taking mean of the z-distributions across channels.

        By default it computes anomaly maps for all the scales as it gave better performance on initial tests.
        Use ``AnomalyMapMode.MAX`` for the largest scale as mentioned in the paper.

        Args:
            inputs (Tensor): z-distributions for the three scales.
            mode (AnomalyMapMode): Anomaly map mode.

        Returns:
            Tensor: Anomaly maps.
        """
        anomaly_map: Tensor
        if self.mode == AnomalyMapMode.ALL:
            anomaly_map = torch.ones(inputs[0].shape[0], 1, *self.input_dims[1:]).to(inputs[0].device)
            for z_dist in inputs:
                mean_z = (z_dist**2).mean(dim=1, keepdim=True)
                anomaly_map *= F.interpolate(
                    mean_z,
                    size=self.input_dims[1:],
                    mode="bilinear",
                    align_corners=False,
                )
        else:
            mean_z = (inputs[0] ** 2).mean(dim=1, keepdim=True)
            anomaly_map = F.interpolate(
                mean_z,
                size=self.input_dims[1:],
                mode="bilinear",
                align_corners=False,
            )

        return anomaly_map



"""Test This script performs inference on the test dataset and saves the output visualizations into a directory."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from argparse import ArgumentParser, Namespace

from pytorch_lightning import Trainer, seed_everything

from anomalib.config import get_configurable_parameters
from anomalib.data import get_datamodule
from anomalib.models import get_model
from anomalib.utils.callbacks import get_callbacks


def get_parser() -> ArgumentParser:
    """Get parser.

    Returns:
        ArgumentParser: The parser object.
    """
    parser = ArgumentParser()
    parser.add_argument("--model", type=str, default="stfpm", help="Name of the algorithm to train/test")
    parser.add_argument("--config", type=str, required=False, help="Path to a model config file")
    parser.add_argument("--weight_file", type=str, default="weights/model.ckpt")

    return parser


def test(args: Namespace):
    """Test an anomaly model.

    Args:
        args (Namespace): The arguments from the command line.
    """
    config = get_configurable_parameters(
        model_name=args.model,
        config_path=args.config,
        weight_file=args.weight_file,
    )

    if config.project.seed:
        seed_everything(config.project.seed)

    datamodule = get_datamodule(config)
    model = get_model(config)

    callbacks = get_callbacks(config)

    trainer = Trainer(callbacks=callbacks, **config.trainer)
    trainer.test(model=model, datamodule=datamodule)


if __name__ == "__main__":
    args = get_parser().parse_args()
    test(args)



"""Region-based Anomaly Detection with Real Time Training and Analysis.

Feature Extractor.
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import torch
from torch import Tensor, nn
from torchvision.ops import RoIAlign
from torchvision.transforms import Normalize, Resize

from anomalib.data.utils.boxes import scale_boxes

WEIGHTS_URL = "https://github.com/openvinotoolkit/anomalib/releases/download/rkde-weights/rkde_feature_extractor.pth"


class FeatureExtractor(nn.Module):
    """Feature Extractor module for Region-based anomaly detection."""

    def __init__(self) -> None:
        super().__init__()

        self.transform = nn.Sequential(
            Resize(size=600, max_size=1000),
            Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        )

        self.features = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=11, stride=4, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(64, 192, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
        )

        self.roi_align = RoIAlign(output_size=(6, 6), spatial_scale=1 / 16, sampling_ratio=0)

        self.classifier = nn.Sequential(
            nn.Linear(256 * 6 * 6, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
        )

        # load the pre-trained weights from url
        self.load_state_dict(torch.hub.load_state_dict_from_url(WEIGHTS_URL, progress=False))

    @torch.no_grad()
    def forward(self, batch: Tensor, rois: Tensor) -> Tensor:
        """Forward pass of the feature extractor.

        Args:
            batch (Tensor): Batch of input images of shape [B, C, H, W]
            rois (Tensor): Tensor of shape [N, 5] describing the regions-of-interest in the batch
        Returns:
            Tensor: Tensor containing a 4096-dimensional feature vector for every RoI location.
        """

        # Apply the feature extractor transforms
        transformed_batch = self.transform(batch)

        # Scale the RoIs to the effective input size of the feature extractor.
        rois[:, 1:] = scale_boxes(rois[:, 1:], batch.shape[-2:], transformed_batch.shape[-2:])

        # Forward pass through the backbone
        features = self.features(transformed_batch)
        features = self.roi_align(features, rois)
        features = torch.flatten(features, 1)
        features = self.classifier(features)
        return features



"""Test test.py entrypoint script."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import sys
from importlib.util import find_spec

import pytest

sys.path.append("tools")
from unittest.mock import patch


@pytest.mark.order(2)
class TestTestEntrypoint:
    """This tests whether the entrypoints run without errors without quantitative measure of the outputs."""

    @pytest.fixture
    def get_functions(self):
        """Get functions from test.py"""
        if find_spec("test") is not None:
            from tools.test import get_parser, test
        else:
            raise Exception("Unable to import test.py for testing")
        return get_parser, test

    def test_test(self, get_functions, get_config, project_path):
        """Test test.py"""
        get_parser, test = get_functions
        with patch("tools.test.get_configurable_parameters", side_effect=get_config):
            # Test when model key is passed
            arguments = get_parser().parse_args(
                ["--model", "padim", "--weight_file", project_path + "/weights/lightning/model.ckpt"]
            )
            test(arguments)

            # Test when weight file is incorrect
            arguments = get_parser().parse_args(["--model", "padim"])
            with pytest.raises(FileNotFoundError):
                test(arguments)

            # Test when config key is passed
            arguments = get_parser().parse_args(
                [
                    "--config",
                    "src/anomalib/models/padim/config.yaml",
                    "--weight_file",
                    project_path + "/weights/lightning/model.ckpt",
                ]
            )
            test(arguments)



"""Regions extraction module of AI-VAD model implementation."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import torch
from torch import Tensor, nn
from torchvision.models.detection import MaskRCNN_ResNet50_FPN_V2_Weights, maskrcnn_resnet50_fpn_v2
from torchvision.ops import box_area, clip_boxes_to_image
from torchvision.transforms.functional import gaussian_blur, rgb_to_grayscale

from anomalib.data.utils.boxes import boxes_to_masks, masks_to_boxes

PERSON_LABEL = 1


class RegionExtractor(nn.Module):
    """Region extractor for AI-VAD.

    Args:
        box_score_thresh (float): Confidence threshold for bounding box predictions.
        persons_only (bool): When enabled, only regions labeled as person are included.
        min_bbox_area (int): Minimum bounding box area. Regions with a surface area lower than this value are excluded.
        max_bbox_overlap (float): Maximum allowed overlap between bounding boxes.
        enable_foreground_detections (bool): Add additional foreground detections based on pixel difference between
            consecutive frames.
        foreground_kernel_size (int): Gaussian kernel size used in foreground detection.
        foreground_binary_threshold (int): Value between 0 and 255 which acts as binary threshold in foreground
            detection.
    """

    def __init__(
        self,
        box_score_thresh: float = 0.8,
        persons_only: bool = False,
        min_bbox_area: int = 100,
        max_bbox_overlap: float = 0.65,
        enable_foreground_detections: bool = True,
        foreground_kernel_size: int = 3,
        foreground_binary_threshold: int = 18,
    ) -> None:
        super().__init__()

        self.persons_only = persons_only
        self.min_bbox_area = min_bbox_area
        self.max_bbox_overlap = max_bbox_overlap
        self.enable_foreground_detections = enable_foreground_detections
        self.foreground_kernel_size = foreground_kernel_size
        self.foreground_binary_threshold = foreground_binary_threshold

        weights = MaskRCNN_ResNet50_FPN_V2_Weights.DEFAULT
        self.backbone = maskrcnn_resnet50_fpn_v2(weights=weights, box_score_thresh=box_score_thresh, rpn_nms_thresh=0.3)

    def forward(self, first_frame: Tensor, last_frame: Tensor) -> list[dict]:
        """Forward pass through region extractor.

        Args:
            first_frame (Tensor): Batch of input images of shape (N, C, H, W) forming the first frames in the clip
            last_frame (Tensor): Batch of input images of shape (N, C, H, W) forming the last frame in the clip
        Returns:
            list[dict]: List of Mask RCNN predictions for each image in the batch.
        """
        with torch.no_grad():
            regions = self.backbone(last_frame)

        if self.enable_foreground_detections:
            regions = self.add_foreground_boxes(
                regions, first_frame, last_frame, self.foreground_kernel_size, self.foreground_binary_threshold
            )

        regions = self.post_process_bbox_detections(regions)

        return regions

    def add_foreground_boxes(
        self,
        regions: list[dict[str, Tensor]],
        first_frame: Tensor,
        last_frame: Tensor,
        kernel_size: int,
        binary_threshold: int,
    ) -> list[dict[str, Tensor]]:
        """Add any foreground regions that were not detected by the region extractor.

        This method adds regions that likely belong to the foreground of the video scene, but were not detected by the
        region extractor module. The foreground pixels are determined by taking the pixel difference between two
        consecutive video frames and applying a binary threshold. The final detections consist of all connected
        components in the foreground that do not fall in one of the bounding boxes predicted by the region extractor.

        Args:
            regions (list[dict[str, Tensor]]): Region detections for a batch of images, generated by the region
                extraction module.
            first_frame (Tensor): video frame at time t-1
            last_frame (Tensor): Video frame time t
            kernel_size (int): Kernel size for Gaussian smoothing applied to input frames
            binary_threshold (int): Binary threshold used in foreground detection, should be in range [0, 255]

        Returns:
            list[dict[str, Tensor]]: region detections with foreground regions appended
        """
        # apply gaussian blur to first and last frame
        first_frame = gaussian_blur(first_frame, [kernel_size, kernel_size])
        last_frame = gaussian_blur(last_frame, [kernel_size, kernel_size])

        # take the abs diff between the blurred images and convert to grayscale
        pixel_diff = torch.abs(first_frame - last_frame)
        pixel_diff = rgb_to_grayscale(pixel_diff).squeeze(1)

        # apply binary threshold to the diff
        foreground_map = (pixel_diff > binary_threshold / 255).int()

        # remove regions already detected by region extractor
        boxes_list = [im_regions["boxes"] for im_regions in regions]
        boxes_list = [
            clip_boxes_to_image(boxes + Tensor([-2, -2, 2, 2]).to(boxes.device), foreground_map.shape[-2:])
            for boxes in boxes_list
        ]  # extend boxes by 2 in all directions to ensure full object is included
        boxes_mask = boxes_to_masks(boxes_list, foreground_map.shape[-2:]).int()
        foreground_map *= -boxes_mask + 1  # invert mask

        # find boxes from foreground map
        batch_boxes, _ = masks_to_boxes(foreground_map)

        # append foreground detections to region extractor detections
        for im_regions, boxes, pixel_mask in zip(regions, batch_boxes, foreground_map):
            if boxes.shape[0] == 0:
                continue

            # append boxes, labels and scores
            im_regions["boxes"] = torch.cat([im_regions["boxes"], boxes])
            im_regions["labels"] = torch.cat(
                [im_regions["labels"], torch.zeros(boxes.shape[0], device=boxes.device)]
            )  # set label as background, in accordance with region extractor predictions
            im_regions["scores"] = torch.cat(
                [im_regions["scores"], torch.ones(boxes.shape[0], device=boxes.device) * 0.5]
            )  # set confidence to 0.5

            # append masks
            im_boxes_as_list = [box.unsqueeze(0) for box in boxes]  # list with one box per element
            boxes_mask = boxes_to_masks(im_boxes_as_list, pixel_mask.shape[-2:]).int()
            new_masks = pixel_mask.repeat((len(im_boxes_as_list), 1, 1)) * boxes_mask
            im_regions["masks"] = torch.cat([im_regions["masks"], new_masks.unsqueeze(1)])

        return regions

    def post_process_bbox_detections(self, regions: list[dict[str, Tensor]]) -> list[dict[str, Tensor]]:
        """Post-process the region detections.

        The region detections are filtered based on class label, bbox area and overlap with other regions.

        Args:
            regions (list[dict[str, Tensor]]): Region detections for a batch of images, generated by the region
                extraction module.

        Returns:
            list[dict[str, Tensor]]: Filtered regions
        """
        filtered_regions = []
        for im_regions in regions:
            if self.persons_only:
                im_regions = self._keep_only_persons(im_regions)
            im_regions = self._filter_by_area(im_regions, self.min_bbox_area)
            im_regions = self._delete_overlapping_boxes(im_regions, self.max_bbox_overlap)
            filtered_regions.append(im_regions)
        return filtered_regions

    def _keep_only_persons(self, regions: dict[str, Tensor]) -> dict[str, Tensor]:
        """Remove all region detections that are not labeled as a person by the region extractor.

        Args:
            regions (dict[str, Tensor]): Region detections for a single image in the batch.

        Returns:
            dict[str, Tensor]: Region detections from which non-person objects have been removed.
        """
        keep = torch.where(regions["labels"] == PERSON_LABEL)
        return self.subsample_regions(regions, keep)

    def _filter_by_area(self, regions: dict[str, Tensor], min_area) -> dict[str, Tensor]:
        """Remove all regions with a surface area smaller than the specified value.

        Args:
            regions (dict[str, Tensor]): Region detections for a single image in the batch.

        Returns:
            dict[str, Tensor]: Region detections from which small regions have been removed.
        """
        areas = box_area(regions["boxes"])
        keep = torch.where(areas > min_area)
        return self.subsample_regions(regions, keep)

    def _delete_overlapping_boxes(self, regions: dict[str, Tensor], threshold: float) -> dict[str, Tensor]:
        """Delete overlapping bounding boxes.

        For each bounding box, the overlap with all other bounding boxes relative to their own surface area is computed.
        When the relative overlap with any other box is higher than the specified threshold, the box is removed. when
        both boxes have a relative overlap higher than the threshold, only the smaller box is removed.

        Args:
            regions (dict[str, Tensor]): Region detections for a single image in the batch.

        Returns:
            dict[str, Tensor]: Region detections from which overlapping regions have been removed.
        """

        # sort boxes by area
        areas = box_area(regions["boxes"])
        indices = areas.argsort()

        keep = []
        for idx in range(len(indices)):
            overlap_coords = torch.hstack(
                [
                    torch.max(regions["boxes"][indices[idx], :2], regions["boxes"][indices[idx + 1 :], :2]),  # x1, y1
                    torch.min(regions["boxes"][indices[idx], 2:], regions["boxes"][indices[idx + 1 :], 2:]),  # x2, y2
                ]
            )
            mask = torch.all(overlap_coords[:, :2] < overlap_coords[:, 2:], dim=1)  # filter non-overlapping
            overlap = box_area(overlap_coords) * mask.int()
            overlap_ratio = overlap / areas[indices[idx]]

            if not any(overlap_ratio > threshold):
                keep.append(indices[idx])

        return self.subsample_regions(regions, torch.stack(keep))

    @staticmethod
    def subsample_regions(regions: dict[str, Tensor], indices: Tensor) -> dict[str, Tensor]:
        """Helper method that subsamples the items in a region dictionary based on a Tensor of indices.

        Args:
            regions (dict[str, Tensor]): Region detections for a single image in the batch.
            indices (Tensor): Indices of region detections that should be kept.

        Returns:
            dict[str, Tensor]: Subsampled region detections.
        """
        new_regions_dict = {}
        for key, value in regions.items():
            new_regions_dict[key] = value[indices]
        return new_regions_dict



"""Fixtures for the tools tests."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from tempfile import TemporaryDirectory
from typing import Generator

import albumentations as A
import cv2
import numpy as np
import pytest
from albumentations.pytorch import ToTensorV2

from anomalib.config import get_configurable_parameters


@pytest.fixture(scope="package")
def project_path() -> Generator[str, None, None]:
    with TemporaryDirectory() as project_dir:
        yield project_dir


@pytest.fixture(scope="package")
def get_config(project_path):
    def get_config(
        model_name: str | None = None,
        config_path: str | None = None,
        weight_file: str | None = None,
    ):
        """Gets config for testing."""
        config = get_configurable_parameters(model_name, config_path, weight_file)
        config.dataset.image_size = (100, 100)
        config.model.input_size = (100, 100)
        config.project.path = project_path
        config.trainer.max_epochs = 1
        config.trainer.check_val_every_n_epoch = 1
        config.trainer.limit_train_batches = 1
        config.trainer.limit_predict_batches = 1
        return config

    yield get_config


@pytest.fixture(scope="package")
def get_dummy_inference_image(project_path) -> Generator[str, None, None]:
    image = np.zeros((100, 100, 3), dtype=np.uint8)
    cv2.imwrite(project_path + "/dummy_image.png", image)
    yield project_path + "/dummy_image.png"


@pytest.fixture(scope="package")
def transforms_config() -> dict:
    """Note: this is computed using trainer.datamodule.test_data.transform.to_dict()"""
    return A.Compose([A.ToFloat(max_value=255), ToTensorV2()]).to_dict()



"""Helper to show progress bars with `urlretrieve`, check hash of file."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import hashlib
import io
import logging
import os
import re
import tarfile
from dataclasses import dataclass
from pathlib import Path
from tarfile import TarFile, TarInfo
from typing import Iterable
from urllib.request import urlretrieve
from zipfile import ZipFile

from tqdm import tqdm

logger = logging.getLogger(__name__)


@dataclass
class DownloadInfo:
    """Info needed to download a dataset from a url."""

    name: str
    url: str
    hash: str
    filename: str | None = None


class DownloadProgressBar(tqdm):
    """Create progress bar for urlretrieve. Subclasses `tqdm`.

    For information about the parameters in constructor, refer to `tqdm`'s documentation.

    Args:
        iterable (Iterable | None): Iterable to decorate with a progressbar.
                            Leave blank to manually manage the updates.
        desc (str | None): Prefix for the progressbar.
        total (int | float | None): The number of expected iterations. If unspecified,
                                            len(iterable) is used if possible. If float("inf") or as a last
                                            resort, only basic progress statistics are displayed
                                            (no ETA, no progressbar).
                                            If `gui` is True and this parameter needs subsequent updating,
                                            specify an initial arbitrary large positive number,
                                            e.g. 9e9.
        leave (bool | None): upon termination of iteration. If `None`, will leave only if `position` is `0`.
        file (io.TextIOWrapper |  io.StringIO | None): Specifies where to output the progress messages
                                                            (default: sys.stderr). Uses `file.write(str)` and
                                                            `file.flush()` methods.  For encoding, see
                                                            `write_bytes`.
        ncols (int | None): The width of the entire output message. If specified,
                            dynamically resizes the progressbar to stay within this bound.
                            If unspecified, attempts to use environment width. The
                            fallback is a meter width of 10 and no limit for the counter and
                            statistics. If 0, will not print any meter (only stats).
        mininterval (float | None): Minimum progress display update interval [default: 0.1] seconds.
        maxinterval (float | None): Maximum progress display update interval [default: 10] seconds.
                                    Automatically adjusts `miniters` to correspond to `mininterval`
                                    after long display update lag. Only works if `dynamic_miniters`
                                    or monitor thread is enabled.
        miniters (int | float | None): Minimum progress display update interval, in iterations.
                                            If 0 and `dynamic_miniters`, will automatically adjust to equal
                                            `mininterval` (more CPU efficient, good for tight loops).
                                            If > 0, will skip display of specified number of iterations.
                                            Tweak this and `mininterval` to get very efficient loops.
                                            If your progress is erratic with both fast and slow iterations
                                            (network, skipping items, etc) you should set miniters=1.
        use_ascii (str | bool | None): If unspecified or False, use unicode (smooth blocks) to fill
                                        the meter. The fallback is to use ASCII characters " 123456789#".
        disable (bool | None): Whether to disable the entire progressbar wrapper
                                    [default: False]. If set to None, disable on non-TTY.
        unit (str | None): String that will be used to define the unit of each iteration
                            [default: it].
        unit_scale (int | float | bool): If 1 or True, the number of iterations will be reduced/scaled
                            automatically and a metric prefix following the
                            International System of Units standard will be added
                            (kilo, mega, etc.) [default: False]. If any other non-zero
                            number, will scale `total` and `n`.
        dynamic_ncols (bool | None): If set, constantly alters `ncols` and `nrows` to the
                                        environment (allowing for window resizes) [default: False].
        smoothing (float | None): Exponential moving average smoothing factor for speed estimates
                                    (ignored in GUI mode). Ranges from 0 (average speed) to 1
                                    (current/instantaneous speed) [default: 0.3].
        bar_format (str | None):  Specify a custom bar string formatting. May impact performance.
                                    [default: '{l_bar}{bar}{r_bar}'], where
                                    l_bar='{desc}: {percentage:3.0f}%|' and
                                    r_bar='| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, '
                                    '{rate_fmt}{postfix}]'
                                    Possible vars: l_bar, bar, r_bar, n, n_fmt, total, total_fmt,
                                    percentage, elapsed, elapsed_s, ncols, nrows, desc, unit,
                                    rate, rate_fmt, rate_noinv, rate_noinv_fmt,
                                    rate_inv, rate_inv_fmt, postfix, unit_divisor,
                                    remaining, remaining_s, eta.
                                    Note that a trailing ": " is automatically removed after {desc}
                                    if the latter is empty.
        initial (int | float | None): The initial counter value. Useful when restarting a progress
                                            bar [default: 0]. If using float, consider specifying `{n:.3f}`
                                            or similar in `bar_format`, or specifying `unit_scale`.
        position (int | None): Specify the line offset to print this bar (starting from 0)
                                    Automatic if unspecified.
                                    Useful to manage multiple bars at once (eg, from threads).
        postfix (dict | None): Specify additional stats to display at the end of the bar.
                                    Calls `set_postfix(**postfix)` if possible (dict).
        unit_divisor (float | None): [default: 1000], ignored unless `unit_scale` is True.
        write_bytes (bool | None): If (default: None) and `file` is unspecified,
                                    bytes will be written in Python 2. If `True` will also write
                                    bytes. In all other cases will default to unicode.
        lock_args (tuple | None): Passed to `refresh` for intermediate output
                                    (initialisation, iterating, and updating).
                                    nrows (int | None): The screen height. If specified, hides nested bars
                                    outside this bound. If unspecified, attempts to use environment height.
                                    The fallback is 20.
        colour (str | None): Bar colour (e.g. 'green', '#00ff00').
        delay (float | None): Don't display until [default: 0] seconds have elapsed.
        gui (bool | None): WARNING: internal parameter - do not use.
                                Use tqdm.gui.tqdm(...) instead. If set, will attempt to use
                                matplotlib animations for a graphical output [default: False].


    Example:
        >>> with DownloadProgressBar(unit='B', unit_scale=True, miniters=1, desc=url.split('/')[-1]) as p_bar:
        >>>         urllib.request.urlretrieve(url, filename=output_path, reporthook=p_bar.update_to)
    """

    def __init__(
        self,
        iterable: Iterable | None = None,
        desc: str | None = None,
        total: int | float | None = None,
        leave: bool | None = True,
        file: io.TextIOWrapper | io.StringIO | None = None,
        ncols: int | None = None,
        mininterval: float | None = 0.1,
        maxinterval: float | None = 10.0,
        miniters: int | float | None = None,
        use_ascii: bool | str | None = None,
        disable: bool | None = False,
        unit: str | None = "it",
        unit_scale: bool | int | float | None = False,
        dynamic_ncols: bool | None = False,
        smoothing: float | None = 0.3,
        bar_format: str | None = None,
        initial: int | float | None = 0,
        position: int | None = None,
        postfix: dict | None = None,
        unit_divisor: float | None = 1000,
        write_bytes: bool | None = None,
        lock_args: tuple | None = None,
        nrows: int | None = None,
        colour: str | None = None,
        delay: float | None = 0,
        gui: bool | None = False,
        **kwargs,
    ):
        super().__init__(
            iterable=iterable,
            desc=desc,
            total=total,
            leave=leave,
            file=file,
            ncols=ncols,
            mininterval=mininterval,
            maxinterval=maxinterval,
            miniters=miniters,
            ascii=use_ascii,
            disable=disable,
            unit=unit,
            unit_scale=unit_scale,
            dynamic_ncols=dynamic_ncols,
            smoothing=smoothing,
            bar_format=bar_format,
            initial=initial,
            position=position,
            postfix=postfix,
            unit_divisor=unit_divisor,
            write_bytes=write_bytes,
            lock_args=lock_args,
            nrows=nrows,
            colour=colour,
            delay=delay,
            gui=gui,
            **kwargs,
        )
        self.total: int | float | None

    def update_to(self, chunk_number: int = 1, max_chunk_size: int = 1, total_size=None) -> None:
        """Progress bar hook for tqdm.

        Based on https://stackoverflow.com/a/53877507
        The implementor does not have to bother about passing parameters to this as it gets them from urlretrieve.
        However the context needs a few parameters. Refer to the example.

        Args:
            chunk_number (int, optional): The current chunk being processed. Defaults to 1.
            max_chunk_size (int, optional): Maximum size of each chunk. Defaults to 1.
            total_size ([type], optional): Total download size. Defaults to None.
        """
        if total_size is not None:
            self.total = total_size
        self.update(chunk_number * max_chunk_size - self.n)


def is_file_potentially_dangerous(file_name: str) -> bool:
    """Check if a file is potentially dangerous.

    Args:
        file_name (str): Filename.

    Returns:
        bool: True if the member is potentially dangerous, False otherwise.

    """
    # Some example criteria. We could expand this.
    unsafe_patterns = ["/etc/", "/root/"]
    for pattern in unsafe_patterns:
        if re.search(pattern, file_name):
            return True
    return False


def safe_extract(tar_file: TarFile, root: Path, members: list[TarInfo]) -> None:
    """Extract safe members from a tar archive.

    Args:
        tar_file (TarFile): TarFile object.
        root (Path): Root directory where the dataset will be stored.
        members (List[TarInfo]): List of safe members to be extracted.

    """
    for member in members:
        tar_file.extract(member, root)


def hash_check(file_path: Path, expected_hash: str) -> None:
    """Raise assert error if hash does not match the calculated hash of the file.

    Args:
        file_path (Path): Path to file.
        expected_hash (str): Expected hash of the file.
    """
    with file_path.open("rb") as hash_file:
        assert (
            hashlib.new(name="md5", data=hash_file.read(), usedforsecurity=False).hexdigest() == expected_hash
        ), f"Downloaded file {file_path} does not match the required hash."


def extract(file_name: Path, root: Path) -> None:
    """Extract a dataset

    Args:
        file_name (Path): Path of the file to be extracted.
        root (Path): Root directory where the dataset will be stored.

    """
    logger.info("Extracting dataset into root folder.")

    # Safely extract zip files
    if file_name.suffix == ".zip":
        with ZipFile(file_name, "r") as zip_file:
            for file_info in zip_file.infolist():
                if not is_file_potentially_dangerous(file_info.filename):
                    zip_file.extract(file_info, root)

    # Safely extract tar files.
    elif file_name.suffix in (".tar", ".gz", ".xz", ".tgz"):
        with tarfile.open(file_name) as tar_file:
            members = tar_file.getmembers()
            safe_members = [member for member in members if not is_file_potentially_dangerous(member.name)]
            safe_extract(tar_file, root, safe_members)

    else:
        raise ValueError(f"Unrecognized file format: {file_name}")

    logger.info("Cleaning up files.")
    file_name.unlink()


def download_and_extract(root: Path, info: DownloadInfo) -> None:
    """Download and extract a dataset.

    Args:
        root (Path): Root directory where the dataset will be stored.
        info (DownloadInfo): Info needed to download the dataset.
    """
    root.mkdir(parents=True, exist_ok=True)

    # save the compressed file in the specified root directory, using the same file name as on the server
    if info.filename:
        downloaded_file_path = root / info.filename
    else:
        downloaded_file_path = root / info.url.split("/")[-1]

    if downloaded_file_path.exists():
        logger.info("Existing dataset archive found. Skipping download stage.")
    else:
        logger.info("Downloading the %s dataset.", info.name)
        with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=info.name) as progress_bar:
            urlretrieve(  # nosec - suppress bandit warning (urls are hardcoded)
                url=f"{info.url}",
                filename=downloaded_file_path,
                reporthook=progress_bar.update_to,
            )
        logger.info("Checking the hash of the downloaded file.")
        hash_check(downloaded_file_path, info.hash)

    extract(downloaded_file_path, root)


def is_within_directory(directory: Path, target: Path):
    """Checks if a target path is located within a given directory.

    Args:
        directory (Path): path of the parent directory
        target (Path): path of the target
    Returns:
        (bool): True if the target is within the directory, False otherwise
    """
    abs_directory = directory.resolve()
    abs_target = target.resolve()

    # TODO: replace with pathlib is_relative_to after switching to Python 3.10
    prefix = os.path.commonprefix([abs_directory, abs_target])
    return prefix == str(abs_directory)



# mypy: ignore-errors
# ruff: noqa

# Original Code
# https://github.com/openai/CLIP.
# SPDX-License-Identifier: MIT
#
# Modified
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import hashlib
import logging
import os
import warnings
from typing import List, Union
from urllib.parse import urlparse

import requests
import torch
from PIL import Image
from pkg_resources import packaging
from torchvision.transforms import CenterCrop, Compose, Normalize, Resize, ToTensor
from tqdm import tqdm

logger = logging.getLogger(__name__)
from .model import build_model

try:
    from torchvision.transforms import InterpolationMode

    BICUBIC = InterpolationMode.BICUBIC
except ImportError:
    BICUBIC = Image.BICUBIC


if packaging.version.parse(torch.__version__) < packaging.version.parse("1.7.1"):
    warnings.warn("PyTorch version 1.7.1 or higher is recommended")


__all__ = ["available_models", "load"]

_MODELS = {
    "RN50": "https://openaipublic.azureedge.net/clip/models/afeb0e10f9e5a86da6080e35cf09123aca3b358a0c3e3b6c78a7b63bc04b6762/RN50.pt",
    "RN101": "https://openaipublic.azureedge.net/clip/models/8fa8567bab74a42d41c5915025a8e4538c3bdbe8804a470a72f30b0d94fab599/RN101.pt",
    "RN50x4": "https://openaipublic.azureedge.net/clip/models/7e526bd135e493cef0776de27d5f42653e6b4c8bf9e0f653bb11773263205fdd/RN50x4.pt",
    "RN50x16": "https://openaipublic.azureedge.net/clip/models/52378b407f34354e150460fe41077663dd5b39c54cd0bfd2b27167a4a06ec9aa/RN50x16.pt",
    "RN50x64": "https://openaipublic.azureedge.net/clip/models/be1cfb55d75a9666199fb2206c106743da0f6468c9d327f3e0d0a543a9919d9c/RN50x64.pt",
    "ViT-B/32": "https://openaipublic.azureedge.net/clip/models/40d365715913c9da98579312b702a82c18be219cc2a73407c4526f58eba950af/ViT-B-32.pt",
    "ViT-B/16": "https://openaipublic.azureedge.net/clip/models/5806e77cd80f8b59890b7e101eabd078d9fb84e6937f9e85e4ecb61988df416f/ViT-B-16.pt",
    "ViT-L/14": "https://openaipublic.azureedge.net/clip/models/b8cca3fd41ae0c99ba7e8951adf17d267cdb84cd88be6f7c2e0eca1737a03836/ViT-L-14.pt",
    "ViT-L/14@336px": "https://openaipublic.azureedge.net/clip/models/3035c92b350959924f9f00213499208652fc7ea050643e8b385c2dac08641f02/ViT-L-14-336px.pt",
}


def _verify_checksum(file_path: str, url: str) -> bool:
    expected_sha256 = url.split("/")[-2]
    sha256_hash = hashlib.sha256()

    with open(file_path, "rb") as file:
        for chunk in iter(lambda: file.read(4096), b""):
            sha256_hash.update(chunk)

    file_hash = sha256_hash.hexdigest()

    return file_hash == expected_sha256


def _download(url: str, root: str):
    os.makedirs(root, exist_ok=True)
    filename = os.path.basename(urlparse(url).path)
    download_target = os.path.join(root, filename)

    if os.path.exists(download_target):
        if not os.path.isfile(download_target):
            raise FileExistsError(f"{download_target} exists and is not a regular file")
        if _verify_checksum(download_target, url):
            return download_target

        logger.warning("%s exists, but the checksum does not match; re-downloading the file", download_target)
        os.remove(download_target)

    response = requests.get(url, stream=True, timeout=10.0)  # Timeout is for bandit security linter
    response.raise_for_status()

    total_size = int(response.headers.get("Content-Length", 0))

    with open(download_target, "wb") as file, tqdm(
        total=total_size, ncols=80, unit="iB", unit_scale=True, unit_divisor=1024
    ) as loop:
        for chunk in response.iter_content(chunk_size=8192):
            if chunk:
                file.write(chunk)
                loop.update(len(chunk))

    if not _verify_checksum(download_target, url):
        raise RuntimeError("Model has been downloaded but the checksum does not match")

    return download_target


def _convert_image_to_rgb(image):
    return image.convert("RGB")


def _transform(n_px):
    return Compose(
        [
            Resize(n_px, interpolation=BICUBIC),
            CenterCrop(n_px),
            _convert_image_to_rgb,
            ToTensor(),
            Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)),
        ]
    )


def available_models() -> List[str]:
    """Returns the names of available CLIP models"""
    return list(_MODELS.keys())


def load(
    name: str,
    device: Union[str, torch.device] = "cuda" if torch.cuda.is_available() else "cpu",
    jit: bool = False,
    download_root: str = None,
):
    """Load a CLIP model

    Parameters
    ----------
    name : str
        A model name listed by `clip.available_models()`, or the path to a model checkpoint containing the state_dict

    device : Union[str, torch.device]
        The device to put the loaded model

    jit : bool
        Whether to load the optimized JIT model or more hackable non-JIT model (default).

    download_root: str
        path to download the model files; by default, it uses "~/.cache/clip"

    Returns
    -------
    model : torch.nn.Module
        The CLIP model

    preprocess : Callable[[PIL.Image], torch.Tensor]
        A torchvision transform that converts a PIL image into a tensor that the returned model can take as its input
    """
    if name in _MODELS:
        model_path = _download(_MODELS[name], download_root or os.path.expanduser("~/.cache/clip"))
    elif os.path.isfile(name):
        model_path = name
    else:
        raise RuntimeError(f"Model {name} not found; available models = {available_models()}")

    with open(model_path, "rb") as opened_file:
        try:
            # loading JIT archive
            model = torch.jit.load(opened_file, map_location=device if jit else "cpu").eval()
            state_dict = None
        except RuntimeError:
            # loading saved state dict
            if jit:
                warnings.warn(f"File {model_path} is not a JIT archive. Loading as a state dict instead")
                jit = False
            state_dict = torch.load(opened_file, map_location="cpu")

    if not jit:
        model = build_model(state_dict or model.state_dict()).to(device)
        if str(device) == "cpu":
            model.float()
        return model, _transform(model.visual.input_resolution)

    # patch the device names
    device_holder = torch.jit.trace(lambda: torch.ones([]).to(torch.device(device)), example_inputs=[])
    device_node = [n for n in device_holder.graph.findAllNodes("prim::Constant") if "Device" in repr(n)][-1]

    def patch_device(module):
        try:
            graphs = [module.graph] if hasattr(module, "graph") else []
        except RuntimeError:
            graphs = []

        if hasattr(module, "forward1"):
            graphs.append(module.forward1.graph)

        for graph in graphs:
            for node in graph.findAllNodes("prim::Constant"):
                if "value" in node.attributeNames() and str(node["value"]).startswith("cuda"):
                    node.copyAttributes(device_node)

    model.apply(patch_device)
    patch_device(model.encode_image)
    patch_device(model.encode_text)

    # patch dtype to float32 on CPU
    if str(device) == "cpu":
        float_holder = torch.jit.trace(lambda: torch.ones([]).float(), example_inputs=[])
        float_input = list(float_holder.graph.findNode("aten::to").inputs())[1]
        float_node = float_input.node()

        def patch_float(module):
            try:
                graphs = [module.graph] if hasattr(module, "graph") else []
            except RuntimeError:
                graphs = []

            if hasattr(module, "forward1"):
                graphs.append(module.forward1.graph)

            for graph in graphs:
                for node in graph.findAllNodes("aten::to"):
                    inputs = list(node.inputs())
                    for i in [1, 2]:  # dtype can be the second or third argument to aten::to()
                        if inputs[i].node()["value"] == 5:
                            inputs[i].node().copyAttributes(float_node)

        model.apply(patch_float)
        patch_float(model.encode_image)
        patch_float(model.encode_text)

        model.float()

    return model, _transform(model.input_resolution.item())



# REMARK: CODE WAS TAKEN FROM https://github.com/eliahuhorwitz/3D-ADS/blob/main/utils/au_pro_util.py

"""
Code based on the official MVTec 3D-AD evaluation code found at
https://www.mydrive.ch/shares/45924/9ce7a138c69bbd4c8d648b72151f839d/download/428846918-1643297332/evaluation_code.tar.xz
Utility functions that compute a PRO curve and its definite integral, given
pairs of anomaly and ground truth maps.
The PRO curve can also be integrated up to a constant integration limit.
"""
from bisect import bisect

import numpy as np
from scipy.ndimage.measurements import label


class GroundTruthComponent:
    """
    Stores sorted anomaly scores of a single ground truth component.
    Used to efficiently compute the region overlap for many increasing thresholds.
    """

    def __init__(self, anomaly_scores):
        """
        Initialize the module.
        Args:
            anomaly_scores: List of all anomaly scores within the ground truth
                            component as numpy array.
        """
        # Keep a sorted list of all anomaly scores within the component.
        self.anomaly_scores = anomaly_scores.copy()
        self.anomaly_scores.sort()

        # Pointer to the anomaly score where the current threshold divides the component into OK / NOK pixels.
        self.index = 0

        # The last evaluated threshold.
        self.last_threshold = None

    def compute_overlap(self, threshold):
        """
        Compute the region overlap for a specific threshold.
        Thresholds must be passed in increasing order.
        Args:
            threshold: Threshold to compute the region overlap.
        Returns:
            Region overlap for the specified threshold.
        """
        if self.last_threshold is not None:
            assert self.last_threshold <= threshold

        # Increase the index until it points to an anomaly score that is just above the specified threshold.
        while self.index < len(self.anomaly_scores) and self.anomaly_scores[self.index] <= threshold:
            self.index += 1

        # Compute the fraction of component pixels that are correctly segmented as anomalous.
        return 1.0 - self.index / len(self.anomaly_scores)


def trapezoid(x, y, x_max=None):
    """
    This function calculates the definit integral of a curve given by x- and corresponding y-values.
    In contrast to, e.g., 'numpy.trapz()', this function allows to define an upper bound to the integration range by
    setting a value x_max.
    Points that do not have a finite x or y value will be ignored with a warning.
    Args:
        x:     Samples from the domain of the function to integrate need to be sorted in ascending order. May contain
               the same value multiple times. In that case, the order of the corresponding y values will affect the
               integration with the trapezoidal rule.
        y:     Values of the function corresponding to x values.
        x_max: Upper limit of the integration. The y value at max_x will be determined by interpolating between its
               neighbors. Must not lie outside of the range of x.
    Returns:
        Area under the curve.
    """

    x = np.array(x)
    y = np.array(y)
    finite_mask = np.logical_and(np.isfinite(x), np.isfinite(y))
    if not finite_mask.all():
        print(
            "WARNING: Not all x and y values passed to trapezoid are finite. Will continue with only the finite values."
        )
    x = x[finite_mask]
    y = y[finite_mask]

    # Introduce a correction term if max_x is not an element of x.
    correction = 0.0
    if x_max is not None:
        if x_max not in x:
            # Get the insertion index that would keep x sorted after np.insert(x, ins, x_max).
            ins = bisect(x, x_max)
            # x_max must be between the minimum and the maximum, so the insertion_point cannot be zero or len(x).
            assert 0 < ins < len(x)

            # Calculate the correction term which is the integral between the last x[ins-1] and x_max. Since we do not
            # know the exact value of y at x_max, we interpolate between y[ins] and y[ins-1].
            y_interp = y[ins - 1] + ((y[ins] - y[ins - 1]) * (x_max - x[ins - 1]) / (x[ins] - x[ins - 1]))
            correction = 0.5 * (y_interp + y[ins - 1]) * (x_max - x[ins - 1])

        # Cut off at x_max.
        mask = x <= x_max
        x = x[mask]
        y = y[mask]

    # Return area under the curve using the trapezoidal rule.
    return np.sum(0.5 * (y[1:] + y[:-1]) * (x[1:] - x[:-1])) + correction


def collect_anomaly_scores(anomaly_maps, ground_truth_maps):
    """
    Extract anomaly scores for each ground truth connected component as well as anomaly scores for each potential false
    positive pixel from anomaly maps.

    Args:
        anomaly_maps: List of anomaly maps (2D numpy arrays) that contain a real-valued anomaly score at each pixel.
        ground_truth_maps: List of ground truth maps (2D numpy arrays) that contain binary-valued ground truth labels
                           for each pixel. 0 indicates that a pixel is anomaly-free. 1 indicates that a pixel contains
                           an anomaly.
    Returns:
        ground_truth_components: A list of all ground truth connected components that appear in the dataset.
            For each component, a sorted list of its anomaly scores is stored.
        anomaly_scores_ok_pixels: A sorted list of anomaly scores of all anomaly-free pixels of the dataset.
            This list can be used to quickly select thresholds that fix a certain false positive rate.
    """
    # Make sure an anomaly map is present for each ground truth map.
    assert len(anomaly_maps) == len(ground_truth_maps)

    # Initialize ground truth components and scores of potential fp pixels.
    ground_truth_components = []
    anomaly_scores_ok_pixels = np.zeros(len(ground_truth_maps) * ground_truth_maps[0].size)

    # Structuring element for computing connected components.
    structure = np.ones((3, 3), dtype=int)

    # Collect anomaly scores within each ground truth region and for all potential fp pixels.
    ok_index = 0
    for gt_map, prediction in zip(ground_truth_maps, anomaly_maps):
        # Compute the connected components in the ground truth map.
        labeled, n_components = label(gt_map, structure)

        # Store all potential fp scores.
        num_ok_pixels = len(prediction[labeled == 0])
        anomaly_scores_ok_pixels[ok_index : ok_index + num_ok_pixels] = prediction[labeled == 0].copy()
        ok_index += num_ok_pixels

        # Fetch anomaly scores within each GT component.
        for k in range(n_components):
            component_scores = prediction[labeled == (k + 1)]
            ground_truth_components.append(GroundTruthComponent(component_scores))

    # Sort all potential false positive scores.
    anomaly_scores_ok_pixels = np.resize(anomaly_scores_ok_pixels, ok_index)
    anomaly_scores_ok_pixels.sort()

    return ground_truth_components, anomaly_scores_ok_pixels


def compute_pro(anomaly_maps, ground_truth_maps, num_thresholds):
    """
    Compute the PRO curve at equidistant interpolation points for a set of anomaly maps with corresponding ground
    truth maps. The number of interpolation points can be set manually.

    Args:
        anomaly_maps: List of anomaly maps (2D numpy arrays) that contain a real-valued anomaly score at each pixel.
        ground_truth_maps: List of ground truth maps (2D numpy arrays) that contain binary-valued ground truth labels
            for each pixel. 0 indicates that a pixel is anomaly-free. 1 indicates that a pixel contains an anomaly.
        num_thresholds: Number of thresholds to compute the PRO curve.

    Returns:
        fprs: List of false positive rates.
        pros: List of correspoding PRO values.
    """
    # Fetch sorted anomaly scores.
    ground_truth_components, anomaly_scores_ok_pixels = collect_anomaly_scores(anomaly_maps, ground_truth_maps)
    # Select equidistant thresholds.
    threshold_positions = np.linspace(0, len(anomaly_scores_ok_pixels) - 1, num=num_thresholds, dtype=int)

    fprs = [1.0]
    pros = [1.0]
    for pos in threshold_positions:
        threshold = anomaly_scores_ok_pixels[pos]

        # Compute the false positive rate for this threshold.
        fpr = 1.0 - (pos + 1) / len(anomaly_scores_ok_pixels)

        # Compute the PRO value for this threshold.
        pro = 0.0
        for component in ground_truth_components:
            pro += component.compute_overlap(threshold)
        pro /= len(ground_truth_components)

        fprs.append(fpr)
        pros.append(pro)

    # Return (FPR/PRO) pairs in increasing FPR order.
    fprs = fprs[::-1]
    pros = pros[::-1]

    return fprs, pros


def calculate_au_pro(gts, predictions, integration_limit=0.3, num_thresholds=100):
    """
    Compute the area under the PRO curve for a set of ground truth images and corresponding anomaly images.
    Args:
        gts:         List of tensors that contain the ground truth images for a single dataset object.
        predictions: List of tensors containing anomaly images for each ground truth image.
        integration_limit:    Integration limit to use when computing the area under the PRO curve.
        num_thresholds:       Number of thresholds to use to sample the area under the PRO curve.
    Returns:
        au_pro:    Area under the PRO curve computed up to the given integration limit.
        pro_curve: PRO curve values for localization (fpr,pro).
    """
    # Compute the PRO curve.
    pro_curve = compute_pro(anomaly_maps=predictions, ground_truth_maps=gts, num_thresholds=num_thresholds)

    # Compute the area under the PRO curve.
    au_pro = trapezoid(pro_curve[0], pro_curve[1], x_max=integration_limit)
    au_pro /= integration_limit

    # Return the evaluation metrics.
    return au_pro, pro_curve



from pathlib import Path
from typing import Union

import torch
from omegaconf.dictconfig import DictConfig
from omegaconf.listconfig import ListConfig
from torch import nn

from anomalib.models.components import AnomalyModule
from anomalib.utils.callbacks import ImageVisualizerCallback
from tests.helpers.dataset import get_dataset_path
from tests.helpers.metrics import get_metrics


class _DummyAnomalyMapGenerator(nn.Module):
    def __init__(self):
        super().__init__()
        self.input_size = (100, 100)
        self.sigma = 4


class _DummyModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.anomaly_map_generator = _DummyAnomalyMapGenerator()


class DummyModule(AnomalyModule):
    """A dummy model which calls visualizer callback on fake images and masks."""

    def __init__(self, hparams: Union[DictConfig, ListConfig]):
        super().__init__()
        self.model = _DummyModel()
        self.task = "segmentation"
        self.mode = "full"
        self.callbacks = [
            ImageVisualizerCallback(
                task=self.task,
                mode=self.mode,
                image_save_path=hparams.project.path + "/images",
                log_images=True,
                save_images=True,
            )
        ]  # test if this is removed

        self.image_metrics, self.pixel_metrics = get_metrics(hparams)
        self.image_metrics.set_threshold(hparams.model.threshold.image_default)
        self.pixel_metrics.set_threshold(hparams.model.threshold.pixel_default)

    def test_step(self, batch, _):
        """Only used to trigger on_test_epoch_end."""
        self.log(name="loss", value=0.0, prog_bar=True)
        outputs = dict(
            image_path=[Path(get_dataset_path("bottle")) / "broken_large/000.png"],
            image=torch.rand((1, 3, 100, 100)),
            mask=torch.zeros((1, 100, 100)),
            anomaly_maps=torch.ones((1, 100, 100)),
            label=torch.Tensor([0]),
            pred_labels=torch.Tensor([0]),
            pred_masks=torch.zeros((1, 100, 100)),
        )
        return outputs

    def validation_epoch_end(self, outputs):
        return None

    def test_epoch_end(self, outputs):
        return None

    def configure_optimizers(self):
        return None
